from bark_ml.library_wrappers.lib_tf_agents.networks.gnns.graph_network import GraphNetwork
from bark_ml.library_wrappers.lib_tf_agents.networks.gnns.interaction_wrapper import InteractionWrapper

def init_interaction_network(name, params):
  return InteractionWrapper(
    params=params, 
    name=name + "_InteractionNetwork")



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import os
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.matplotlib_viewer import MPViewer
from bark.runtime.scenario.scenario_generation.config_with_ease import \
  LaneCorridorConfig, ConfigWithEase
from bark.core.models.dynamic import SingleTrackModel
from bark.core.world.opendrive import XodrDrivingDirection
from bark.core.world.goal_definition import GoalDefinitionPolygon
from bark.core.world.goal_definition import GoalDefinitionStateLimitsFrenet
from bark.core.models.behavior import BehaviorLaneChangeRuleBased, BehaviorIDMClassic, \
  BehaviorMobilRuleBased

from bark_ml.environments.blueprints.blueprint import Blueprint
from bark_ml.evaluators.goal_reached import GoalReached

from bark_ml.evaluators.goal_reached_guiding import GoalReachedGuiding
from bark_ml.observers.nearest_state_observer import NearestAgentsObserver
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMacroActionsML, \
    BehaviorDiscreteMotionPrimitivesML
from bark_ml.core.observers import NearestObserver



class MergingLaneCorridorConfig(LaneCorridorConfig):
  def __init__(self,
               params=None,
               **kwargs):
    super(MergingLaneCorridorConfig, self).__init__(params, **kwargs)

  def goal(self, world):
    road_corr = world.map.GetRoadCorridor(
      self._road_ids, XodrDrivingDirection.forward)
    lane_corr = self._road_corridor.lane_corridors[0]
    lane_polygon = lane_corr.polygon
    return GoalDefinitionStateLimitsFrenet(lane_corr.center_line,
                                           (0.9, 0.9),
                                           (0.15, 0.15),
                                           (5., 15.))


class MergingBlueprint(Blueprint):
  def __init__(self,
               params=None,
               num_scenarios=250,
               random_seed=0,
               ml_behavior=None,
               viewer=True,
               mode="dense"):
    params["BehaviorIDMClassic"]["BrakeForLaneEnd"] = True
    params["BehaviorIDMClassic"]["BrakeForLaneEndEnabledDistance"] = 100.
    params["BehaviorIDMClassic"]["BrakeForLaneEndDistanceOffset"] = 30.
    params["BehaviorIDMClassic"]["DesiredVelocity"] = 10.
    params["World"]["remove_agents_out_of_map"] = False
    # TODO: modify dense
    # ds_min and ds_max
    if mode == "dense":
      ds_min = 7.
      ds_max = 12.
    if mode == "medium":
      ds_min = 12.
      ds_max = 17.
    left_lane = MergingLaneCorridorConfig(
      params=params,
      road_ids=[0, 1],
      ds_min=ds_min,
      ds_max=ds_max,
      min_vel=9.,
      max_vel=11.,
      s_min=5.,
      s_max=45.,
      lane_corridor_id=0,
      controlled_ids=None)
    right_lane = MergingLaneCorridorConfig(
      params=params,
      road_ids=[0, 1],
      lane_corridor_id=1,
      ds_min=ds_min,
      ds_max=ds_max,
      s_min=5.,
      s_max=25.,
      min_vel=9.,
      max_vel=11.,
      behavior_model=BehaviorMobilRuleBased(params),
      controlled_ids=True)
    scenario_generation = \
      ConfigWithEase(
        num_scenarios=num_scenarios,
        map_file_name=os.path.join(os.path.dirname(__file__), "../../../environments/blueprints/merging/DR_DEU_Merging_MT_v01_centered.xodr"),  # NOLINT
        random_seed=random_seed,
        params=params,
        lane_corridor_configs=[left_lane, right_lane])
    if viewer:
      viewer = MPViewer(params=params,
                        x_range=[-20, 20],
                        y_range=[-20, 20],
                        follow_agent_id=True)
    dt = 0.2
    # evaluator = GoalReachedGuiding(params)
    evaluator = GoalReached(params)
    observer = NearestObserver(params)
    ml_behavior = ml_behavior

    super().__init__(
      scenario_generation=scenario_generation,
      viewer=viewer,
      dt=dt,
      evaluator=evaluator,
      observer=observer,
      ml_behavior=ml_behavior)


class ContinuousMergingBlueprint(MergingBlueprint):
  def __init__(self,
               params=None,
               num_scenarios=25,
               random_seed=0,
               viewer=True,
               mode="dense"):
    ml_behavior = BehaviorContinuousML(params)
    MergingBlueprint.__init__(self,
                              params=params,
                              num_scenarios=num_scenarios,
                              random_seed=random_seed,
                              ml_behavior=ml_behavior,
                              viewer=viewer,
                              mode=mode)


class DiscreteMergingBlueprint(MergingBlueprint):
  def __init__(self,
               params=None,
               num_scenarios=25,
               random_seed=0,
               viewer=True,
               mode="dense"):
    ml_behavior = BehaviorDiscreteMacroActionsML(params)
    MergingBlueprint.__init__(self,
                              params=params,
                              num_scenarios=num_scenarios,
                              random_seed=random_seed,
                              ml_behavior=ml_behavior,
                              viewer=viewer,
                              mode=mode)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from bark.core.world.agent import Agent
from bark.core.world import World
from bark.core.world.map import MapInterface
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.commons.xodr_parser import XodrParser
import copy
import os
from pathlib import Path

# Module variable to maintain map directory
__MAPFILE_DIRECTORY = None


def SetMapfileDirectory(dir):
    global __MAPFILE_DIRECTORY
    __MAPFILE_DIRECTORY = dir


def GetMapfileDirectory():
    global __MAPFILE_DIRECTORY
    return __MAPFILE_DIRECTORY


class Scenario:
    def __init__(self,
                 agent_list=None,
                 eval_agent_ids=None,
                 map_file_name=None,
                 json_params=None,
                 map_interface=None,
                 observer_model=None):
        self._agent_list = agent_list or []
        self._eval_agent_ids = eval_agent_ids or []
        self._map_file_name = map_file_name
        self._json_params = json_params
        self._map_interface = map_interface
        self._observer_model = observer_model

    @property
    def map_file_name(self):
        return self._map_file_name

    @property
    def full_map_file_name(self):
        if GetMapfileDirectory():
            return os.path.join(GetMapfileDirectory(), self._map_file_name)
        return self._map_file_name

    @property
    def json_params(self):
        return self._json_params

    @property
    def eval_agent_ids(self):
        return self._eval_agent_ids

    @property
    def map_interface(self):
        return self._map_interface

    @map_interface.setter
    def map_interface(self, map_interface):
        self._map_interface = map_interface

    def GetWorldState(self):
        """get initial world state of scenario to start simulation from here

        Returns:
            [bark.core.world.World]
        """
        return self._build_world_state()

    def copy(self):
        observer_model = None
        if self._observer_model is not None:
          observer_model = copy.deepcopy(self._observer_model)
        return Scenario(agent_list=copy.deepcopy(self._agent_list),
                        eval_agent_ids=self._eval_agent_ids.copy(),
                        map_file_name=self._map_file_name,
                        json_params=self._json_params.copy(),
                        map_interface=self._map_interface,
                        observer_model=observer_model)

    def _build_world_state(self):
        param_server = ParameterServer(json=self._json_params)
        world = World(param_server)
        if self._observer_model is not None:
          world.SetObserverModel(self._observer_model)
        if self._map_interface is None:
            self.CreateMapInterface(self.full_map_file_name)
            world.SetMap(self._map_interface)
        else:
            world.SetMap(self._map_interface)
        for agent in self._agent_list:
            agent.GenerateRoadCorridor(self._map_interface)
            world.AddAgent(agent)
        world.UpdateAgentRTree()
        return world

    def __getstate__(self):
        odict = self.__dict__.copy()
        del odict['_map_interface']
        return odict

    def __setstate__(self, sdict):
        sdict['_map_interface'] = None
        self.__dict__.update(sdict)

    # TODO(@hart): should be a commons function
    def CreateMapInterface(self, map_file_name):
        map_file_load_test = Path(map_file_name)
        if map_file_load_test.is_file():
            xodr_parser = XodrParser(map_file_name)
        else:
            print("Searching for map file {}".format(map_file_name))
            objects_found = sorted(Path().rglob(map_file_name))
            if len(objects_found) == 0:
                raise ValueError("No Map found")
            elif len(objects_found) > 1:
                raise ValueError("Multiple Maps found")
            else:
                xodr_parser = XodrParser(objects_found[0].as_posix())

        map_interface = MapInterface()
        map_interface.SetOpenDriveMap(xodr_parser.map)
        self._map_interface = map_interface

    def GetDatasetScenarioDescription(self):
        # only relevant for scenarios from dataset
        try:
            track_id_ego = self.eval_agent_ids[0]
            track_file_name = self.json_params["track_file"]
            dataset_scenario_desc = {
                'TrackIdEgo': track_id_ego, 'TrackFileName': track_file_name}
        except:
            dataset_scenario_desc = {}
        return dataset_scenario_desc



# Copyright (c) 2020 Patrick Hart, Julian Bernhard,
# Klemens Esterle, Tobias Kessler
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from bark.core.world.evaluation import \
  EvaluatorGoalReached, EvaluatorCollisionEgoAgent, \
  EvaluatorStepCount, EvaluatorDrivableArea, CaptureAgentStates
from bark.runtime.commons.parameters import ParameterServer

from bark_ml.evaluators.evaluator import StateEvaluator


class GoalReached(StateEvaluator):
  def __init__(self,
               params=ParameterServer(),
               eval_agent=None):
    StateEvaluator.__init__(self, params)
    self._goal_reward = \
      self._params["ML"]["GoalReachedEvaluator"]["GoalReward",
        "Reward for reaching the goal.",
        1.]
    self._col_penalty = \
      self._params["ML"]["GoalReachedEvaluator"]["CollisionPenalty",
        "Reward given for a collisions.",
        -1.]
    self._max_steps = \
      self._params["ML"]["GoalReachedEvaluator"]["MaxSteps",
        "Maximum steps per episode.",
        60]
    self._eval_agent = eval_agent

  def _add_evaluators(self):
    evaluators = {}
    evaluators["goal_reached"] = EvaluatorGoalReached()
    evaluators["collision"] = EvaluatorCollisionEgoAgent()
    evaluators["step_count"] = EvaluatorStepCount()
    evaluators["drivable_area"] = EvaluatorDrivableArea()
    return evaluators

  def _evaluate(self, observed_world, eval_results, action):
    """Returns information about the current world state
    """
    done = False
    success = eval_results["goal_reached"]
    step_count = eval_results["step_count"]
    collision = eval_results["collision"] or eval_results["drivable_area"] or (step_count > self._max_steps)
    # for agent_id, agent in observed_world.agents.items():
    #   eval_results[agent_id] = agent.state
    # determine whether the simulation should terminate
    if success or collision or step_count > self._max_steps:
      done = True
    if collision:
      success = 0
      eval_results["goal_reached"] = 0
      
    # calculate reward
    reward = collision * self._col_penalty + \
      success * self._goal_reward
    return reward, done, eval_results
    
  def Reset(self, world):
    return super(GoalReached, self).Reset(world)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Marco Oliva
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

# bark
from bark.core.models.behavior import BehaviorModel, BehaviorDynamicModel

import tensorflow as tf

from tf_agents.policies import greedy_policy
from tf_agents.agents.sac import sac_agent
from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.utils.common import Checkpointer
from tf_agents.trajectories import time_step as ts

from bark_ml.library_wrappers.lib_tf_agents.networks import GNNActorNetwork, GNNCriticNetwork
from bark_ml.library_wrappers.lib_tf_agents.agents.tfa_agent import BehaviorTFAAgent
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.library_wrappers.lib_tf_agents.agents.gnn_initializers import init_interaction_network
  

class BehaviorGraphSACAgent(BehaviorTFAAgent):
  """
  SAC-Agent with graph neural networks.
  This agent is based on the tf-agents library.
  """
  
  def __init__(self,
               environment=None,
               observer=None,
               params=None,
               init_gnn='init_interaction_network'):
    """
    Initializes a `BehaviorGraphSACAgent` instance.

    Args:
    environment:
    observer: The `GraphObserver` instance that generates the observations.
    params: A `ParameterServer` instance containing parameters to configure
      the agent.
    """
    # the super init calls 'GetAgent', so assign the observer before
    self._gnn_sac_params = params["ML"]["BehaviorGraphSACAgent"]
    self._init_gnn = eval(init_gnn)
    BehaviorTFAAgent.__init__(self,
                              environment=environment,
                              params=params,
                              observer=observer)
    self._replay_buffer = self.GetReplayBuffer()
    self._dataset = self.GetDataset()
    self._collect_policy = self.GetCollectionPolicy()
    self._eval_policy = self.GetEvalPolicy()

  def GetAgent(self, env, params):
    self._params["ML"]["GraphDims"] = self._observer.graph_dimensions
    
    # actor network
    actor_net = GNNActorNetwork(
      input_tensor_spec=env.observation_spec(),
      output_tensor_spec=env.action_spec(),
      gnn=self._init_gnn,
      fc_layer_params=self._gnn_sac_params[
        "ActorFcLayerParams", "", [256, 256]],
      params=params
    )

    # critic network
    critic_net = GNNCriticNetwork(
      (env.observation_spec(), env.action_spec()),
      gnn=self._init_gnn,
      observation_fc_layer_params=self._gnn_sac_params[
        "CriticObservationFcLayerParams", "", [256]],
      action_fc_layer_params=self._gnn_sac_params[
        "CriticActionFcLayerParams", "", None],
      joint_fc_layer_params=self._gnn_sac_params[
        "CriticJointFcLayerParams", "", [256, 256]],
      params=params
    )
    
    # agent
    tf_agent = sac_agent.SacAgent(
      env.time_step_spec(),
      env.action_spec(),
      actor_network=actor_net,
      critic_network=critic_net,
      actor_optimizer=tf.compat.v1.train.AdamOptimizer(
          learning_rate=self._gnn_sac_params["ActorLearningRate", "", 3e-4]),
      critic_optimizer=tf.compat.v1.train.AdamOptimizer(
          learning_rate=self._gnn_sac_params["CriticLearningRate", "", 3e-4]),
      alpha_optimizer=tf.compat.v1.train.AdamOptimizer(
          learning_rate=self._gnn_sac_params["AlphaLearningRate", "", 0.]),
      target_update_tau=self._gnn_sac_params["TargetUpdateTau", "", 1.],
      target_update_period=self._gnn_sac_params["TargetUpdatePeriod", "", 1],
      td_errors_loss_fn=tf.math.squared_difference,
      gamma=self._gnn_sac_params["Gamma", "", 0.995],
      reward_scale_factor=self._gnn_sac_params["RewardScaleFactor", "", 1.],
      train_step_counter=self._ckpt.step,
      name=self._gnn_sac_params["AgentName", "", "gnn_sac_agent"],
      debug_summaries=self._gnn_sac_params["DebugSummaries", "", True])
    tf_agent.initialize()
    return tf_agent

  def GetReplayBuffer(self):
    return tf_uniform_replay_buffer.TFUniformReplayBuffer(
      data_spec=self._agent.collect_data_spec,
      batch_size=self._wrapped_env.batch_size,
      max_length=self._gnn_sac_params["ReplayBufferCapacity", "", 10000])

  def GetDataset(self):
    dataset = self._replay_buffer.as_dataset(
      num_parallel_calls=self._gnn_sac_params["ParallelBufferCalls", "", 1],
      sample_batch_size=self._gnn_sac_params["BatchSize", "", 512],
      num_steps=self._gnn_sac_params["BufferNumSteps", "", 2]) \
        .prefetch(self._gnn_sac_params["BufferPrefetch", "", 3])
    return dataset

  def GetCollectionPolicy(self):
    return self._agent.collect_policy

  def GetEvalPolicy(self):
    return greedy_policy.GreedyPolicy(self._agent.policy)

  def Reset(self):
    pass

  @property
  def collect_policy(self):
    return self._collect_policy

  @property
  def eval_policy(self):
    return self._eval_policy
  



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from bark.runtime.scenario import Scenario
from bark.runtime.scenario.scenario_generation import ScenarioGeneration
from bark.runtime.commons import ModelJsonConversion
from bark.core.world.agent import *
from bark.core.models.behavior import *
from bark.core.world import *
from bark.core.world.goal_definition import GoalDefinition, GoalDefinitionPolygon, GoalDefinitionSequential
from bark.core.world.map import *
from bark.core.models.dynamic import *
from bark.core.models.execution import *
from bark.core.geometry import *
from bark.core.geometry.standard_shapes import *
from bark.runtime.commons import ParameterServer
from bark.runtime.commons import XodrParser

import numpy as np
import math


class DeterministicDroneChallengeGeneration(ScenarioGeneration):
  def __init__(self, num_scenarios, params=None, random_seed=None):
    super(DeterministicDroneChallengeGeneration, self).__init__(params,
                                                                num_scenarios)
    self.initialize_params(params)

  def initialize_params(self, params):
    self._local_params = \
      self._params["Scenario"]["Generation"]["DeterministicDroneChallengeGeneration"]

    self.goal_frame_center = self._local_params["goal_frame"][
      "center_pose", "Center pose of the goal frames",
      [1, 0, 0]]
    self.goal_frame_points = self._local_params["goal_frame"]["polygon_points",
      "Points of the goal frame polygon",
      [[0., 0.],
       [2., 0.],
       [2., 0.5],
       [0., 0.5],
       [0., 0.]]]
    self.goal_frame_poses = self._local_params["goal_poses",
      "A list with x,y, theta specyfing the sequence of the goal frames",
      [[5, 10, -0.3],
      [10,15, -1.2],
      [20,10, -2.9],
      [15,2, -3.9]]]

    self._map_file_name = self._local_params["MapFilename",
     "Path to the open drive map", 
     "bark/runtime/tests/data/Crossing8Course.xodr"]
    self._json_converter = ModelJsonConversion()


  def create_scenarios(self, params, num_scenarios, random_seed):
    """ 
        see baseclass
    """
    scenario_list = []
    for scenario_idx in range(0, num_scenarios):
      scenario = self.create_single_scenario()     
      scenario_list.append(scenario)
    return scenario_list

  def _build_sequential_goal_definition(self):
    goal_list = []
    for goal_pose in self.goal_frame_poses:
        goal_polygon = Polygon2d(self.goal_frame_center,
                               np.array(self.goal_frame_points))
        goal_polygon = goal_polygon.Transform(goal_pose)
        goal_definition = GoalDefinitionPolygon(goal_polygon)
        goal_list.append(goal_definition)
    
    return GoalDefinitionSequential(goal_list)

  def create_single_scenario(self):
    scenario = Scenario(map_file_name=self._map_file_name,
                        json_params=self._params.ConvertToDict())
    scenario._map_interface = None
    world = scenario.GetWorldState()
    agent_list = []
    scenario._agent_list = []
    for agent_json_ in self._local_params["Agents"]:
      agent_json = agent_json_["VehicleModel"].copy()
      agent_json["map_interface"] = world.map
      goal_polygon = Polygon2d([0, 0, 0],
                               np.array(agent_json["goal"]["polygon_points"]))
      goal_polygon = goal_polygon.Translate(Point2d(agent_json["goal"]["center_pose"][0],
                                                    agent_json["goal"]["center_pose"][1]))

      agent_json["goal_definition"] = self._build_sequential_goal_definition()
      agent_state = np.array(agent_json["state"])
      if len(np.shape(agent_state)) > 1:
        agent_state = np.random.uniform(low=agent_state[:, 0],
                                                high=agent_state[:, 1])
      agent_json["state"] = agent_state.tolist()
      agent = self._json_converter.agent_from_json(agent_json,
                                                   param_server=self._local_params)
      agent.SetAgentId(agent_json["id"])
      scenario._agent_list.append(agent)
    scenario._eval_agent_ids = [self._local_params["EgoAgentId",
                                "ID of the ego-agent",
                                0]]
    return scenario


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import os
import shutil
import logging

from bark.runtime.commons.parameters import ParameterServer
from bark.core.commons import FrenetState
from bark.runtime.scenario.interaction_dataset_processing.interaction_dataset_reader import *
from bark.runtime.scenario.interaction_dataset_processing.agent_track_info import AgentTrackInfo
from bark.runtime.scenario.interaction_dataset_processing.scenario_track_info import ScenarioTrackInfo
from bark.runtime.scenario.scenario import Scenario
from bark.core.geometry import Point2d, Collide, Within
from bark.core.geometry.standard_shapes import *

from com_github_interaction_dataset_interaction_dataset.python.utils import dataset_reader


class DatasetDecomposer:
    def __init__(self, map_interface, road_corridor, track_filename, vehicle_length_max, xy_offset, starting_offset_ms=0):
        self._map_interface = map_interface
        self._track_filename = track_filename
        self._starting_offset_ms = starting_offset_ms
        self._xy_offset = xy_offset
        self._vehicle_length_max = vehicle_length_max
        self._road_corridor = road_corridor
        self._track_dict = dataset_reader.read_tracks(track_filename)
        self._agents_track_infos = self.__setup_agents_track_infos__()

    def __find_first_ts_on_map__(self, id_ego):
        traj = TrajectoryFromTrack(
            self._track_dict[id_ego], xy_offset=self._xy_offset)
        for state in traj:
            # wb = WheelbaseFromTrack(self._track_dict[id_ego])
            # r = ColRadiusFromTrack(self._track_dict[id_ego])
            # agent_shape = GenerateCarLimousine(wb, r)
            agent_shape = CarLimousine()
            agent_shape = agent_shape.Transform([state[1], state[2], state[3]])
            for lc in self._road_corridor.lane_corridors:
                frenet_state = FrenetState(state, lc.center_line)
                inside_road = Within(agent_shape, self._road_corridor.polygon)
                aligned_with_road = abs(frenet_state.angle) < 0.1
                if inside_road and aligned_with_road:
                    ms_to_s = 1e3 # scale from s (BARK) to ms (dataset)
                    time_ego_first = state[0] * ms_to_s + self._starting_offset_ms
                    return time_ego_first

        return None

    def __setup_agents_track_infos__(self):
        # dictionary mapping first valid timestamp to agnet id
        agents_track_infos = {}
        for agent_id in self._track_dict.keys():
            # TODO: this could be made optional
            first_ts_on_map = self.__find_first_ts_on_map__(agent_id)
            track = self._track_dict[agent_id]
            if first_ts_on_map is None:
                logging.info("Skip, as agent {} not found on map".format(agent_id))
                pass
            elif str(track.agent_type) != "car":
                logging.info("Skip, as agent {} is of type {}".format(agent_id, track.agent_type))
                pass
            elif track.length > self._vehicle_length_max:
                logging.info("Skip, as agent {} exceeds max length of {} with length {}".format(agent_id, self._vehicle_length_max, track.length))
                pass
            else:
                start_time = first_ts_on_map
                end_time = track.time_stamp_ms_last
                new_agent = AgentTrackInfo(filename=self._track_filename, track_id=agent_id,
                                           start_time=start_time, end_time=end_time)
                agents_track_infos[agent_id] = new_agent
        return agents_track_infos

    def __get_agent_track_info__(self, agent_id):
        if agent_id in self._agents_track_infos.keys():
            return self._agents_track_infos[agent_id]
        else:
            raise ValueError("agent {} not available".format(agent_id))

    def __find_other_ids__(self, id_ego):
        # for each agent extract ids of other agents present in the same time span
        list_ids = []
        time_ego_first = self.__get_agent_track_info__(id_ego).GetStartTimeMs()
        time_ego_last = self.__get_agent_track_info__(id_ego).GetEndTimeMs()

        for id_o in self._agents_track_infos.keys():
            if id_ego == id_o:
                pass
            elif self.__get_agent_track_info__(id_o).GetEndTimeMs() <= time_ego_first:
                # other ends too early
                pass
            elif self.__get_agent_track_info__(id_o).GetStartTimeMs() >= time_ego_last:
                # other starts too late
                pass
            else:
                list_ids.append(id_o)

        return list_ids

    def __collect_all_scenario_infos__(self):
        scenario_list = []
        for id_ego in self._agents_track_infos.keys():
            ego_track_info = self.__get_agent_track_info__(id_ego)
            new_scenario = ScenarioTrackInfo(
                track_filename=self._track_filename, ego_track_info=ego_track_info, xy_offset=self._xy_offset)

            ids_others = self.__find_other_ids__(id_ego)
            for id_o in ids_others:
                agent_o_track_info = self.__get_agent_track_info__(id_o)
                new_scenario.AddTrackInfoOtherAgent(agent_o_track_info)

            scenario_list.append(new_scenario)

        return scenario_list

    def decompose(self):

        scenario_list = self.__collect_all_scenario_infos__()
        return scenario_list



#!/usr/bin/env python

try:
    print("Trying to import csv (for loading track files).")
    import csv
    print("ok")
except ImportError:
    print("Could not import csv, but it's needed to load the track files.")

lanelet2_ok = False
try:
    print("\nTrying to import lanelet2 (for loading maps).")
    import lanelet2
    print("ok")
    lanelet2_ok = True
except ImportError:
    string = "Could not import lanelet2. It must be built and sourced, " + \
             "see https://github.com/fzi-forschungszentrum-informatik/Lanelet2 for details." + \
             "For visualization, the fallback is utils.map_vis_without_lanelet.py."
    print(string)

try:
    print("\nTrying to import xml (for loading maps without lanelet2).")
    import xml.etree.ElementTree as xml
    print("ok")
except ImportError:
    print("Could not import xml.")
    if lanelet2_ok:
        print("  Not too bad, since you have lanelet2 to load the map.")
    else:
        print("  Since lanelet2 also does not work, you need to install this to be able to load and visualize maps.")

try:
    print("\nTrying to import pyproj (for loading maps without lanelet2).")
    import pyproj
    print("ok")
except ImportError:
    print("Could not import pyproj.")
    if lanelet2_ok:
        print("  Not too bad, since you have lanelet2 to load the map.")
    else:
        print("  Since lanelet2 also does not work, you need to install this to be able to load and visualize maps.")


try:
    print("\nTrying to import argparse (argument parsing for loading and visualizing tracks).")
    import argparse
    print("ok")
except ImportError:
    print("Could not import argparse.")

try:
    print("\nTrying to import os (check file existence for loading and visualizing tracks).")
    import os
    print("ok")
except ImportError:
    print("Could not import os.")

try:
    print("\nTrying to import time (for scenario playback).")
    import time
    print("ok")
except ImportError:
    print("Could not import time.")

try:
    print("\nTrying to import math (for visualizing tracks).")
    import math
    print("ok")
except ImportError:
    print("Could not import math.")

try:
    print("\nTrying to import matplotlib (for visualizing tracks).")
    import matplotlib
    print("ok")
except ImportError:
    print("Could not import matplotlib.")

try:
    print("\nTrying to import numpy (for visualizing tracks).")
    import numpy
    print("ok")
except ImportError:
    print("Could not import numpy.")



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import gin
import tensorflow as tf  # pylint: disable=g-explicit-tensorflow-version-import

from tf_agents.networks import network, utils, encoding_network
from bark.runtime.commons.parameters import ParameterServer


@gin.configurable
class GNNValueNetwork(network.Network):
  """Feed Forward value network. Reduces to 1 value output per batch item."""

  def __init__(self,
               input_tensor_spec,
               gnn,
               preprocessing_layers=None,
               preprocessing_combiner=None,
               conv_layer_params=None,
               fc_layer_params=(75, 40),
               dropout_layer_params=None,
               activation_fn=tf.nn.relu,
               kernel_initializer=None,
               batch_squash=False,
               dtype=tf.float32,
               name='ValueNetwork',
               params=ParameterServer()):
    """Creates an instance of `ValueNetwork`.
    Network supports calls with shape outer_rank + observation_spec.shape. Note
    outer_rank must be at least 1.
    Args:
      input_tensor_spec: A `tensor_spec.TensorSpec` or a tuple of specs
        representing the input observations.
      preprocessing_layers: (Optional.) A nest of `tf.keras.layers.Layer`
        representing preprocessing for the different observations.
        All of these layers must not be already built. For more details see
        the documentation of `networks.EncodingNetwork`.
      preprocessing_combiner: (Optional.) A keras layer that takes a flat list
        of tensors and combines them. Good options include
        `tf.keras.layers.Add` and `tf.keras.layers.Concatenate(axis=-1)`.
        This layer must not be already built. For more details see
        the documentation of `networks.EncodingNetwork`.
      conv_layer_params: Optional list of convolution layers parameters, where
        each item is a length-three tuple indicating (filters, kernel_size,
        stride).
      fc_layer_params: Optional list of fully_connected parameters, where each
        item is the number of units in the layer.
      dropout_layer_params: Optional list of dropout layer parameters, each item
        is the fraction of input units to drop or a dictionary of parameters
        according to the keras.Dropout documentation. The additional parameter
        `permanent', if set to True, allows to apply dropout at inference for
        approximated Bayesian inference. The dropout layers are interleaved with
        the fully connected layers; there is a dropout layer after each fully
        connected layer, except if the entry in the list is None. This list must
        have the same length of fc_layer_params, or be None.
      activation_fn: Activation function, e.g. tf.keras.activations.relu,.
      kernel_initializer: Initializer to use for the kernels of the conv and
        dense layers. If none is provided a default variance_scaling_initializer
      batch_squash: If True the outer_ranks of the observation are squashed into
        the batch dimension. This allow encoding networks to be used with
        observations with shape [BxTx...].
      dtype: The dtype to use by the convolution and fully connected layers.
      name: A string representing name of the network.
    Raises:
      ValueError: If input_tensor_spec is not an instance of network.InputSpec.
    """
    super(GNNValueNetwork, self).__init__(
        input_tensor_spec=input_tensor_spec,
        state_spec=(),
        name=name)

    if not kernel_initializer:
      kernel_initializer = tf.compat.v1.keras.initializers.glorot_uniform()
      
    if gnn is None:
      raise ValueError('`gnn` must not be `None`.')

    self._gnn = gnn(name=name + "_GNN", params=params)
    
    self._encoder = encoding_network.EncodingNetwork(
      input_tensor_spec=tf.TensorSpec([None, self._gnn._embedding_size]),
      preprocessing_layers=None,
      preprocessing_combiner=None,
      conv_layer_params=conv_layer_params,
      fc_layer_params=fc_layer_params,
      dropout_layer_params=dropout_layer_params,
      activation_fn=tf.keras.activations.relu,
      kernel_initializer=kernel_initializer,
      batch_squash=False,
      dtype=tf.float32)

    self._postprocessing_layers = tf.keras.layers.Dense(
        1,
        activation=None,
        kernel_initializer=tf.compat.v1.initializers.random_uniform(
            minval=-0.03, maxval=0.03))

  def call(self, observations, step_type=None, network_state=(), training=False):
    # print(observations)
    if len(tf.shape(observations)) == 3:
      observations = tf.squeeze(observations, axis=0)

    embeddings = self._gnn(observations, training=training)
    
    if tf.shape(embeddings)[0] > 0:
      embeddings = embeddings[:, 0] # extract ego state
    
    with tf.name_scope("PPOCriticNetwork"):
      tf.summary.histogram("embedding", embeddings)

    state, network_state = self._encoder(
      embeddings,
      step_type=step_type,
      network_state=network_state,
      training=training)

    value = self._postprocessing_layers(state, training=training)
    value = tf.expand_dims(value, axis=0)
    return tf.squeeze(value, -1), network_state


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import json
import os
from bark.core.commons import Params
import logging

class ParameterServer(Params):
    def __init__(self, **kwargs):
        Params.__init__(self)
        self.param_filename = None
        self.log_if_default = kwargs.pop("log_if_default", False)
        if "filename" in kwargs:
            self.load(kwargs["filename"])
            self.param_filename = kwargs["filename"]
        elif "json" in kwargs:
            self.ConvertToParam(kwargs["json"])
        else:
            self.store = dict()
        self.param_descriptions = dict()
        

    def __getitem__(self, key):
        if isinstance(key, tuple):
            new_key = key[0]
            self.param_descriptions[new_key] = key[1]
            if len(key) == 3:
                default_val = key[2]
            else:
                default_val = "MISSING"
        else:
            new_key = key
        if new_key in self.store:
            return self.store[new_key]
        else:
            if isinstance(key, tuple):  # if key for parameter
                if isinstance(default_val, list) and all(type(el) is dict for el in default_val):
                  value_tmp = []
                  for list_el in default_val:
                    value_tmp.append(ParameterServer(json=list_el))
                  default_val = value_tmp
                self.store[new_key] = default_val
                if self.log_if_default:
                  logging.warning("Using default {} for {}".format(
                  default_val, new_key))
                return self.store[new_key]
            else:  # else it is a Params instance
                self.store[new_key] = ParameterServer(log_if_default = self.log_if_default)
                return self.store[new_key]

    def __contains__(self, key):
        return key in self.store

    def AppendParamServer(self, p_server):
        for key in p_server.store.keys():
            if key in self.store:
                val_self = self.store[key]
                val_other = p_server.store[key]
                if isinstance(val_self, ParameterServer) and isinstance(val_other, ParameterServer):
                    val_self.AppendParamServer(val_other)
                else:
                    logging.warning("Cannot append conflicting key '{}'!".format(key))
            else:
                self.__setitem__(key, p_server[key])

    def FindKey(self, param_key):
      delimiter = "::"
      for key, value in self.store.items():
                  if isinstance(value, ParameterServer):
                      found_key_tmp =  value.FindKey(param_key)
                      if found_key_tmp:
                        return "{}{}{}".format(key, delimiter,
                              found_key_tmp)
                  else:
                      if param_key == key:
                        return param_key
      return None


    def __setitem__(self, key, value):
        store = self.store
        new_key = key
        if isinstance(key, tuple):
            new_key = key[0]
            self.param_descriptions[new_key] = key[1]

        if isinstance(value, list) and all(isinstance(el, dict) for el in value):
          value_tmp = []
          for list_el in value:
            value_tmp.append(ParameterServer(json=list_el))
          value = value_tmp

        if isinstance(key, str):
          self._set_item_from_hierarchy_string(key, value)

    def _set_item_from_hierarchy_string(self, key, value):
      delim = "::"
      found = key.find(delim)
      if found > -1:
        child_params = self.AddChild(key.rsplit(delim, 1)[0])
        new_key = key.rsplit(delim, 1)[1]
        if isinstance(child_params, list):
          for child in child_params:
            child.store[new_key] = value
        else:
          child_params.store[new_key] = value
      else:
        self.store[key] = value

    def __delitem__(self, key):
        if key in self.store:
          del self.store[key]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def __getstate__(self):
        return self.ConvertToDict()

    def __setstate__(self, d):
        self.__init__(json=d)

    def load(self, fn):
        with open(fn) as file:
            dict = json.load(file)
        self.ConvertToParam(dict)

    def ConvertToParam(self, params):
        if isinstance(params, ParameterServer):
          return
        self.store = dict()
        for key, value in params.items():
            if isinstance(value, dict):
                param = ParameterServer(log_if_default = self.log_if_default)
                self.store[key] = param.ConvertToParam(value)
            elif isinstance(value, list) and all(type(el) is dict for el in value):
                param_server_list = []
                for v in value:
                    param_server_list.append(ParameterServer(json=v))
                self.store[key] = param_server_list
            else:
                self.GetValFromString(key, "", value, log_if_default=False)
        return self

    def clone(self):
      return ParameterServer(json = self.ConvertToDict(), \
          log_if_default = self.log_if_default)

    def ConvertToDict(self, print_description=False):
        dict = {}
        for key, value in self.store.items():
            if isinstance(value, ParameterServer):
                v = value.ConvertToDict(print_description)
                if len(v) == 0:
                    if print_description:
                        if key in self.param_descriptions:
                            dict[key] = self.param_descriptions[key]
                        else:
                            dict[key] = "--"
                    else:
                        dict[key] = "MISSING"
                else:
                    dict[key] = v
            elif isinstance(value, list) and all(type(el) is ParameterServer for el in value):
              v_tmp = []
              for v in value:
                v_tmp.append(v.ConvertToDict())
              dict[key] = v_tmp
            else:
                dict[key] = self._ItemToDict(value)
                if print_description:
                    if key in self.param_descriptions:
                        dict[key] = self.param_descriptions[key]
                    else:
                        dict[key] = "--"

        return dict

    def _ItemToDict(self, item):
        if isinstance(item, list):
            ret = []
            for v in item:
                ret.append(self._ItemToDict(v))
        elif isinstance(item, dict):
            ret = {}
            for (key, val) in item.items():
                ret[key] = self._ItemToDict(val)
        elif isinstance(item, ParameterServer):
            ret = item.ConvertToDict()
        else:
            ret = item
        return ret

    def Save(self, filename, print_description=False):
        #if not os.path.exists(os.path.dirname(filename)):
            #try:
           #     os.makedirs(os.path.dirname(filename))
          #  except:
         #       print("Could not dump parameters, no rights to create file directory.")
        #        return
        with open(filename, 'w') as outfile:
            print("Writing parameters to {}".format(os.path.abspath(filename)))
            outfile.write(
                json.dumps(
                    self.ConvertToDict(print_description=print_description),
                    indent=4))

    def get_val_iter(self, hierarchy, description, default_value):
        if not hierarchy:
            raise ValueError("PARAM HIERARCHY IS EMPTY")
        else:
            key = hierarchy[0]
            hierarchy.pop(0)
            if key in self.store:
                #if isinstance(self.store[key], ParameterServer):
                if type(self.store[key]) == ParameterServer:
                    #raise ValueError("get here")
                    return self.store[key].get_val_iter(
                        hierarchy, description, default_value)
                else:
                    return self.store[key], False
            else:
                if not hierarchy:
                    self.store[key] = default_value
                    return default_value, True
                else:
                    self.store[key] = ParameterServer(log_if_default = self.log_if_default)
                    return self.store[key].get_val_iter(
                        hierarchy, description, default_value)
        return

    def GetCondensedParamList(self):
        def check_append(value):
          if isinstance(value, float) or isinstance(value, int) \
              or isinstance(value,bool) or isinstance(value, str):
              return True
          # list double
          elif isinstance(value, list) and all(isinstance(el, float) for el in value):
            return True
          # list list double
          elif isinstance(value, list):
              for el in value:
                if not isinstance(el, list):
                  return False
                for e in el:
                  if not isinstance(e, float):
                    return False
              return True
          return False
        hierarchy_delimiter = "::"
        condensed_param_list = []
        for key, value in self.store.items():
            if isinstance(value, ParameterServer):
                child_list = value.GetCondensedParamList()
                for param_tuple in child_list:
                    param_name = "{}{}{}".format(key,
                                    hierarchy_delimiter, param_tuple[0])
                    param_value = param_tuple[1]
                    if check_append(param_value):
                        condensed_param_list.append((param_name, param_value))
            else:
                if check_append(value):
                    condensed_param_list.append((key, value))
        test = condensed_param_list
        return condensed_param_list

    def GetValFromString(self, hierarchy, description, default_value, log_if_default):
        hierarchy = [x.strip() for x in hierarchy.split("::")]
        if log_if_default:
          # first search for key, otherwise default already integrated into store
          found_key = self.FindKey(hierarchy[-1])
        value, used_default = self.get_val_iter(hierarchy.copy(), description, default_value)
        if log_if_default and used_default:
          logging.warning("Using default {} for {}".format(
                                default_value, hierarchy))
          if found_key:
            logging.warning("Did you mean param {}".format(found_key))
          else:
            logging.warning("Key was nowhere found.")
        return value

    # get values
    def GetBool(self, param_name, description, default_value):
        #return self[param_name, description, default_value]
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def GetReal(self, param_name, description, default_value):
        #return self[param_name, description, default_value]
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def GetInt(self, param_name, description, default_value):
        #return self[param_name, description, default_value]
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def GetListListFloat(self, param_name, description, default_value):
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)
    
    def GetListFloat(self, param_name, description, default_value):
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def GetString(self, param_name, description, default_value):
        return self.GetValFromString(param_name, description, default_value, self.log_if_default)

    def access(self, param_name):
        return self[param_name]

    def SetBool(self, param_name, value):
        self[param_name] = value
        return

    def SetReal(self, param_name, value):
        self[param_name] = value
        return

    def SetInt(self, param_name, value):
        self[param_name] = value
        return

    def AddChild(self, name, delete = False):
        delim = "::"
        rest_name = ""
        child_name = name

        found = name.find(delim)
        if found > -1:
          child_name = name[0:found]
          rest_name = name[found + len(delim):]

        child = None
        if child_name in self.store and not delete:
          child = self.store[child_name]
        else:
          self.store[child_name] = ParameterServer(log_if_default=self.log_if_default)
          child = self.store[child_name]

        if len(rest_name) == 0:
          return child
        else:
          if isinstance(child, list) and all(type(el) is ParameterServer for el in child):
            child_list = []
            for ch in child:
              child_list.append(ch.AddChild(rest_name))
            return child_list
          else:
            return child.AddChild(rest_name)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import os
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.matplotlib_viewer import MPViewer
from bark.runtime.scenario.scenario_generation.configurable_scenario_generation import \
  ConfigurableScenarioGeneration
from bark.core.models.dynamic import SingleTrackModel
from bark.core.world.opendrive import XodrDrivingDirection
from bark.core.world.goal_definition import GoalDefinitionPolygon
from bark.core.world.goal_definition import GoalDefinitionStateLimitsFrenet
from bark.core.models.behavior import BehaviorLaneChangeRuleBased, BehaviorIDMClassic, \
  BehaviorMobilRuleBased

from bark_ml.environments.blueprints.blueprint import Blueprint
from bark_ml.evaluators.goal_reached import GoalReached
from bark_ml.evaluators.goal_reached_guiding import GoalReachedGuiding
from bark_ml.observers.nearest_state_observer import NearestAgentsObserver
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteML
from bark_ml.core.observers import NearestObserver



class ConfigurableScenarioBlueprint(Blueprint):
  def __init__(self,
               params=None,
               ml_behavior=None,
               viewer=True):
    # NOTE: scneario number is wrong
    scenario_generation = ConfigurableScenarioGeneration(
      num_scenarios=2500,
      params=params)
    if viewer:
      viewer = MPViewer(params=params,
                        x_range=[-35, 35],
                        y_range=[-35, 35],
                        follow_agent_id=True)
    dt = 0.2
    # NOTE: evaluator and observer could be overwritten
    evaluator = GoalReached(params)
    observer = NearestObserver(params)
    ml_behavior = ml_behavior
    super().__init__(
      scenario_generation=scenario_generation,
      viewer=viewer,
      dt=dt,
      evaluator=evaluator,
      observer=observer,
      ml_behavior=ml_behavior)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import os
from bark.runtime.commons.parameters import ParameterServer
from bark.runtime.viewer.matplotlib_viewer import MPViewer
from bark.runtime.scenario.scenario_generation.config_with_ease import \
  LaneCorridorConfig, ConfigWithEase
from bark.core.models.dynamic import SingleTrackModel
from bark.core.world.opendrive import XodrDrivingDirection
from bark.core.world.goal_definition import GoalDefinitionPolygon
from bark.core.models.behavior import BehaviorIntersectionRuleBased
from bark.core.world.goal_definition import GoalDefinitionPolygon

from bark_ml.environments.blueprints.blueprint import Blueprint
from bark_ml.evaluators.goal_reached import GoalReached
# from bark_ml.observers.nearest_state_observer import NearestAgentsObserver
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMacroActionsML, \
  BehaviorDiscreteMotionPrimitivesML

from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMacroActionsML, \
  BehaviorDiscreteMotionPrimitivesML
from bark_ml.core.observers import NearestObserver


class IntersectionLaneCorridorConfig(LaneCorridorConfig):
  def __init__(self,
               params=None,
               **kwargs):
    super(IntersectionLaneCorridorConfig, self).__init__(params, **kwargs)

  def controlled_goal(self, world):
    road_corr = world.map.GetRoadCorridor(
      self._road_ids, XodrDrivingDirection.forward)
    lane_corr = self._road_corridor.lane_corridors[0]
    # lanes are sorted by their s-value
    lanes = lane_corr.lanes
    s_val = max(lanes.keys())
    return GoalDefinitionPolygon(lanes[s_val].polygon)


class IntersectionBlueprint(Blueprint):
  def __init__(self,
               params=None,
               num_scenarios=250,
               random_seed=0,
               ml_behavior=None,
               viewer=True):
    lane_corridors = []
    lane_corridors.append(
      IntersectionLaneCorridorConfig(params=params,
                                     source_pos=[-40, -3],
                                     sink_pos=[40, -3],
                                     behavior_model=BehaviorIntersectionRuleBased(params),
                                     min_vel=10.,
                                     max_vel=15.,
                                     ds_min=10.,
                                     ds_max=15.,
                                     s_min=5.,
                                     s_max=50.,
                                     controlled_ids=None))
    lane_corridors.append(
      IntersectionLaneCorridorConfig(params=params,
                                     source_pos=[40, 3],
                                     sink_pos=[-40, 3],
                                     behavior_model=BehaviorIntersectionRuleBased(params),
                                     min_vel=10.,
                                     max_vel=15.,
                                     ds_min=15.,
                                     ds_max=15.,
                                     s_min=5.,
                                     s_max=50.,
                                     controlled_ids=None))
    lane_corridors.append(
      IntersectionLaneCorridorConfig(params=params,
                                     source_pos=[3, -30],
                                     sink_pos=[-40, 3],
                                     behavior_model=BehaviorIntersectionRuleBased(params),
                                     min_vel=5.,
                                     max_vel=10.,
                                     ds_min=10.,
                                     ds_max=15.,
                                     s_min=10.,
                                     s_max=25.,
                                     controlled_ids=True))

    scenario_generation = \
      ConfigWithEase(
        num_scenarios=num_scenarios,
        map_file_name=os.path.join(os.path.dirname(__file__), "../../../environments/blueprints/intersection/4way_intersection.xodr"),  # NOLINT
        random_seed=random_seed,
        params=params,
        lane_corridor_configs=lane_corridors)
    if viewer:
      viewer = MPViewer(params=params,
                        x_range=[-35, 35],
                        y_range=[-35, 35],
                        follow_agent_id=True)
    dt = 0.2
    evaluator = GoalReached(params)
    observer = NearestObserver(params)
    ml_behavior = ml_behavior

    super().__init__(
      scenario_generation=scenario_generation,
      viewer=viewer,
      dt=dt,
      evaluator=evaluator,
      observer=observer,
      ml_behavior=ml_behavior)


class ContinuousIntersectionBlueprint(IntersectionBlueprint):
  def __init__(self,
               params=None,
               num_scenarios=25,
               random_seed=0,
               viewer=True):
    ml_behavior = BehaviorContinuousML(params)
    IntersectionBlueprint.__init__(self,
                                   params=params,
                                   num_scenarios=num_scenarios,
                                   random_seed=random_seed,
                                   ml_behavior=ml_behavior,
                                   viewer=True)


class DiscreteIntersectionBlueprint(IntersectionBlueprint):
  def __init__(self,
               params=None,
               num_scenarios=25,
               random_seed=0):
    ml_behavior = BehaviorDiscreteMacroActionsML(params)
    IntersectionBlueprint.__init__(self,
                                   params=params,
                                   num_scenarios=num_scenarios,
                                   random_seed=random_seed,
                                   ml_behavior=ml_behavior,
                                   viewer=True)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import gym
from gym.envs.registration import register

from bark.runtime.commons.parameters import ParameterServer

from bark_ml.environments.blueprints.highway.highway import \
  ContinuousHighwayBlueprint, DiscreteHighwayBlueprint
from bark_ml.environments.blueprints.merging.merging import \
  ContinuousMergingBlueprint, DiscreteMergingBlueprint
from bark_ml.environments.blueprints.intersection.intersection import \
  ContinuousIntersectionBlueprint, DiscreteIntersectionBlueprint
from bark_ml.environments.single_agent_runtime import SingleAgentRuntime


# highway
class ContinuousHighwayGym(SingleAgentRuntime, gym.Env):
  def __init__(self):
    params = ParameterServer()
    cont_highway_bp = ContinuousHighwayBlueprint(params)
    SingleAgentRuntime.__init__(self,
      blueprint=cont_highway_bp, render=True)

class DiscreteHighwayGym(SingleAgentRuntime, gym.Env):
  def __init__(self, params = ParameterServer(), render=False):
    discrete_highway_bp = DiscreteHighwayBlueprint(params)
    SingleAgentRuntime.__init__(self,
      blueprint=discrete_highway_bp, render=render)

# merging
class ContinuousMergingGym(SingleAgentRuntime, gym.Env):
  def __init__(self):
    params = ParameterServer()
    cont_merging_bp = ContinuousMergingBlueprint(params)
    SingleAgentRuntime.__init__(self,
      blueprint=cont_merging_bp, render=True)

class DiscreteMergingGym(SingleAgentRuntime, gym.Env):
  def __init__(self, params = ParameterServer(), render=False):
    discrete_merging_bp = DiscreteMergingBlueprint(params)
    SingleAgentRuntime.__init__(self,
      blueprint=discrete_merging_bp, render=render)

# merging
class MediumContinuousMergingGym(SingleAgentRuntime, gym.Env):
  def __init__(self):
    params = ParameterServer()
    cont_merging_bp = ContinuousMergingBlueprint(params, mode="medium")
    SingleAgentRuntime.__init__(self,
      blueprint=cont_merging_bp,
      render=True)

# intersection
class ContinuousIntersectionGym(SingleAgentRuntime, gym.Env):
  def __init__(self):
    params = ParameterServer()
    cont_merging_bp = ContinuousIntersectionBlueprint(params)
    SingleAgentRuntime.__init__(self,
      blueprint=cont_merging_bp, render=True)

class DiscreteIntersectionGym(SingleAgentRuntime, gym.Env):
  def __init__(self, params = ParameterServer(), render=False):
    discrete_merging_bp = DiscreteIntersectionBlueprint(params)
    SingleAgentRuntime.__init__(self,
      blueprint=discrete_merging_bp, render=render)

class GymSingleAgentRuntime(SingleAgentRuntime, gym.Wrapper):
    def __init__(self, *args, **kwargs):
        SingleAgentRuntime.__init__(self, *args, **kwargs)


# register gym envs
register(
  id='highway-v0',
  entry_point='bark_ml.environments.gym:ContinuousHighwayGym'
)
register(
  id='highway-v1',
  entry_point='bark_ml.environments.gym:DiscreteHighwayGym'
)
register(
  id='merging-v0',
  entry_point='bark_ml.environments.gym:ContinuousMergingGym'
)
register(
  id='merging-medium-v0',
  entry_point='bark_ml.environments.gym:MediumContinuousMergingGym'
)
register(
  id='merging-v1',
  entry_point='bark_ml.environments.gym:DiscreteMergingGym'
)
register(
  id='intersection-v0',
  entry_point='bark_ml.environments.gym:ContinuousIntersectionGym'
)
register(
  id='intersection-v1',
  entry_point='bark_ml.environments.gym:DiscreteIntersectionGym'
)


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from gym import spaces
import numpy as np
from bark.core.models.dynamic import StateDefinition
from bark.core.world import World, ObservedWorld
from bark.runtime.commons.parameters import ParameterServer
import math
import operator
import functools
import tensorflow as tf

from bark.core.geometry import Distance, Point2d
from bark_ml.observers.observer import StateObserver


class GraphObserverV2(StateObserver):
  def __init__(self, params=ParameterServer()):
    StateObserver.__init__(self, params)
    self._filter_values = [int(StateDefinition.X_POSITION),
                           int(StateDefinition.Y_POSITION),
                           int(StateDefinition.THETA_POSITION),
                            int(StateDefinition.VEL_POSITION)]

    self._self_con = False
    self._state_definition = 250
    self.graph_dimensions = None

  def _position(self, agent) -> Point2d:
    return Point2d(
      agent.state[int(StateDefinition.X_POSITION)],
      agent.state[int(StateDefinition.Y_POSITION)]
    )
    
  def _agents_sorted_by_distance(self, ego_agent, agents):
    """
    Returns the given list of `agents`, sorted in ascending
    order by their relative distance to the `ego_agent`.
    """
    def distance(agent):
      return Distance(
        self._position(ego_agent), 
        self._position(agent))
    if len(agents) > 2:
      agents.sort(key=distance)
    return agents
  
  def GetNearbyAgents(self, center_agent, agents, radius: float, include_ego=False):
    center_agent_pos = self._position(center_agent)
    if include_ego == False:
      other_agents = filter(lambda a: a.id != center_agent.id, agents.values())
    else:
      other_agents = agents.values()
    nearby_agents = []
    for agent in other_agents:
      agent_pos = self._position(agent)
      distance = Distance(center_agent_pos, agent_pos)
      if distance <= radius:
        nearby_agents.append(agent)
    return self._agents_sorted_by_distance(center_agent, nearby_agents)

  def GenSortedList(self, observed_world, radius=150, max_agents=5):
    agent_by_node_pos = {}
    nearby_agents = self.GetNearbyAgents(
      observed_world.ego_agent,
      observed_world.other_agents,
      radius)
    agent_by_node_pos[0] = observed_world.ego_agent
    insertion_idx = 1
    for agent in nearby_agents:
      agent_by_node_pos[insertion_idx] = agent
      insertion_idx += 1
      if insertion_idx == max_agents:
        break
    return agent_by_node_pos
    
  def QueryNodePos(self, agent_map, agent):
    return list(agent_map.keys())[list(agent_map.values()).index(agent)]
  
  def CalcEdgeValue(self, other_agent, agent):
    return self._norm(agent.state) - self._norm(other_agent.state)
  
  def Observe(self, observed_world):
    """see base class
    """
    pos_agent = self.GenSortedList(observed_world)
    node_vals = []
    edge_vals = []
    edge_index = []
    for node_pos, agent in sorted(pos_agent.items()):
      # TODO: make dynamic
      node_vals.append(self._norm(agent.state)[1:])
      # TODO: make dynamic
      nearby_agents = self.GetNearbyAgents(agent, pos_agent, 1000, include_ego=True)
      for nearby_agent in nearby_agents:
        nearby_agent_node_pos = self.QueryNodePos(pos_agent, nearby_agent)
        if self._self_con == False and nearby_agent_node_pos == node_pos:
          continue
        # TODO: make dynamic
        edge_value = self.CalcEdgeValue(nearby_agent, agent)[1:]
        edge_vals.append(edge_value)
        edge_ids = [nearby_agent_node_pos, node_pos]
        edge_index.append(edge_ids)
    
    node_vals = np.reshape(np.vstack(node_vals), (1, -1))
    edge_vals = np.reshape(np.vstack(edge_vals), (1, -1))
    edge_index = np.reshape(np.vstack(edge_index), (1, -1))
    observation = np.concatenate([
      np.array([[np.shape(node_vals)[1]]], dtype=np.float32),
      np.array([[np.shape(edge_vals)[1]]], dtype=np.float32),
      node_vals, edge_vals, edge_index], axis=1)
    
    # equality transf
    obs = np.zeros(shape=(1, self._len_ego_state))
    obs[0, :np.shape(observation)[1]] = observation
    
    obs = tf.convert_to_tensor(obs, dtype=tf.float32)
    return tf.reshape(obs, [-1])
  
  @classmethod
  def graph(cls, observations, graph_dims=None, dense=False):
    batch_size = tf.shape(observations)[0]
    node_vals = []
    edge_vals = []
    edge_indices = []
    node_lens = []
    edge_lens = []
    globals = []
    edge_idx_start = 0
    for obs in observations:
      len_nodes = tf.cast(obs[0], dtype=tf.int32)
      len_edges = tf.cast(obs[1], dtype=tf.int32)
      # TODO: make dynamic
      node_val = tf.reshape(obs[2:len_nodes+2], [-1, 4])
      # TODO: make dynamic
      edge_val = tf.reshape(obs[len_nodes+2:len_nodes+2+len_edges], [-1, 4])
      # print(len_edges)
      edge_index = tf.cast(
        tf.reshape(
          obs[len_nodes+2+len_edges:(len_nodes+2+len_edges + tf.cast(len_edges/2, dtype=tf.int32))], [-1, 2]), dtype=tf.int32) + edge_idx_start
      node_vals.append(node_val)
      edge_indices.append(edge_index)
      edge_vals.append(edge_val)
      edge_idx_start += tf.shape(node_val)[0]
      node_lens.append([tf.shape(node_val)[0]])
      edge_lens.append([tf.shape(edge_val)[0]])
      # globals.append(node_val[0, :])
    
    node_vals = tf.concat(node_vals, axis=0)
    edge_indices = tf.concat(edge_indices, axis=0)
    edge_vals = tf.concat(edge_vals, axis=0)
    node_lens = tf.cast(tf.concat(node_lens, axis=0), dtype=tf.int32)
    edge_lens = tf.cast(tf.concat(edge_lens, axis=0), dtype=tf.int32)
    globals = tf.stack(globals, axis=0)
    return node_vals, edge_indices, node_lens, edge_lens, globals, edge_vals
    
  @property
  def observation_space(self):
    return spaces.Box(
      low=np.zeros(self._len_ego_state),
      high=np.ones(self._len_ego_state))

  def _norm(self, agent_state):
    if not self._normalization_enabled:
      return agent_state
    agent_state[int(StateDefinition.X_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.X_POSITION)],
                          self._world_x_range)
    agent_state[int(StateDefinition.Y_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.Y_POSITION)],
                          self._world_y_range)
    agent_state[int(StateDefinition.THETA_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.THETA_POSITION)],
                          self._theta_range)
    agent_state[int(StateDefinition.VEL_POSITION)] = \
      self._norm_to_range(agent_state[int(StateDefinition.VEL_POSITION)],
                          self._velocity_range)
    print(agent_state)
    return agent_state

  def _norm_to_range(self, value, range):
    return (value - range[0])/(range[1]-range[0])

  @property
  def _len_ego_state(self):
    return self._state_definition


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Marco Oliva
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

# bark
from bark.core.models.behavior import BehaviorModel, BehaviorDynamicModel

import tensorflow as tf

from tf_agents.policies import greedy_policy
from tf_agents.agents.ppo import ppo_agent
from tf_agents.agents.ppo import ppo_clip_agent
from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.utils.common import Checkpointer
from tf_agents.trajectories import time_step as ts

from bark_ml.library_wrappers.lib_tf_agents.networks import GNNActorDistributionNetwork, GNNValueNetwork
from bark_ml.library_wrappers.lib_tf_agents.agents.tfa_agent import BehaviorTFAAgent
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.library_wrappers.lib_tf_agents.agents.gnn_initializers import init_interaction_network


class BehaviorGraphPPOAgent(BehaviorTFAAgent):
  """
  ppo-Agent with graph neural networks.
  This agent is based on the tf-agents library.
  """
  
  def __init__(self,
               environment=None,
               observer=None,
               params=None,
               init_gnn='init_gat'):
    """
    Initializes a `BehaviorGraphppoAgent` instance.

    Args:
    environment:
    observer: The `GraphObserver` instance that generates the observations.
    params: A `ParameterServer` instance containing parameters to configure
      the agent.
    """
    # the super init calls 'GetAgent', so assign the observer before
    self._gnn_ppo_params = params["ML"]["BehaviorGraphPPOAgent"]
    self._init_gnn = eval(init_gnn)
    BehaviorTFAAgent.__init__(self,
                              environment=environment,
                              params=params,
                              observer=observer)
    self._replay_buffer = self.GetReplayBuffer()
    self._collect_policy = self.GetCollectionPolicy()
    self._eval_policy = self.GetEvalPolicy()

  def GetAgent(self, env, params):
    self._params["ML"]["GraphDims"] = self._observer.graph_dimensions
    # actor network
    actor_net = GNNActorDistributionNetwork(
      input_tensor_spec=env.observation_spec(),
      output_tensor_spec=env.action_spec(),
      gnn=self._init_gnn,
      fc_layer_params=self._gnn_ppo_params[
        "ActorFcLayerParams", "", [512, 256, 256]],
      params=params
    )

    # critic network
    value_net = GNNValueNetwork(
      env.observation_spec(),
      gnn=self._init_gnn,
      fc_layer_params=tuple(self._gnn_ppo_params[
        "CriticFcLayerParams", "", [512, 256, 256]]),
      params=params
    )
    
    # agent
    tf_agent = ppo_agent.PPOAgent(
      env.time_step_spec(),
      env.action_spec(),
      optimizer=tf.compat.v1.train.AdamOptimizer(learning_rate=3e-4),
      actor_net=actor_net,
      value_net=value_net,
      normalize_observations=False,
      normalize_rewards=False,
      use_gae=False,
      debug_summaries=True,
      summarize_grads_and_vars=False,
      train_step_counter=self._ckpt.step)
    
    tf_agent.initialize()
    return tf_agent

  def GetReplayBuffer(self):
    return tf_uniform_replay_buffer.TFUniformReplayBuffer(
      data_spec=self._agent.collect_data_spec,
      batch_size=self._wrapped_env.batch_size,
      max_length=self._gnn_ppo_params["ReplayBufferCapacity", "", 10000])

  def GetCollectionPolicy(self):
    return self._agent.collect_policy

  def GetEvalPolicy(self):
    return self._agent.policy

  def Reset(self):
    pass

  @property
  def collect_policy(self):
    return self._collect_policy

  @property
  def eval_policy(self):
    return self._eval_policy


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import unittest
import numpy as np
import os
import matplotlib
import time

from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark.runtime.commons.parameters import ParameterServer
from bark_ml.environments.blueprints import ContinuousHighwayBlueprint, DiscreteHighwayBlueprint
from bark_ml.environments.single_agent_runtime import SingleAgentRuntime
from bark_ml.evaluators.goal_reached import GoalReached
from bark_ml.core.evaluators import GoalReachedEvaluator


class PyEvaluatorTests(unittest.TestCase):
  def test_goal_reached_evaluator(self):
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params)
    env = SingleAgentRuntime(blueprint=bp, render=True)
    env.reset()
    world = env._world

    eval_id = env._scenario._eval_agent_ids[0]
    observed_world = world.Observe([eval_id])[0]
    evaluator = GoalReached(params)
    action = np.array([0., 0.], dtype=np.float32)
    start_time = time.time()
    print(evaluator.Evaluate(observed_world, action))
    end_time = time.time()
    print(f"It took {end_time-start_time} seconds.")
    

  def test_goal_reached_cpp_evaluator(self):
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params)
    env = SingleAgentRuntime(blueprint=bp, render=True)
    env.reset()
    world = env._world

    eval_id = env._scenario._eval_agent_ids[0]
    observed_world = world.Observe([eval_id])[0]
    evaluator = GoalReachedEvaluator(params)
    action = np.array([0., 0.], dtype=np.float32)
    start_time = time.time()
    print(evaluator.Evaluate(observed_world, action))
    end_time = time.time()
    print(f"It took {end_time-start_time} seconds.")

  def test_reward_shaping_evaluator(self):
    params = ParameterServer()
    bp = ContinuousHighwayBlueprint(params)
    env = SingleAgentRuntime(blueprint=bp, render=True)
    env.reset()
    world = env._world

    eval_id = env._scenario._eval_agent_ids[0]
    observed_world = world.Observe([eval_id])[0]
    evaluator = GoalReachedEvaluator(params)
    action = np.array([0., 0.], dtype=np.float32)
    start_time = time.time()
    print(evaluator.Evaluate(observed_world, action))
    end_time = time.time()
    print(f"It took {end_time-start_time} seconds.")
  
if __name__ == '__main__':
  unittest.main()


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import tensorflow as tf

# tfa
from tf_agents.networks import actor_distribution_network
from tf_agents.networks import actor_distribution_rnn_network
from tf_agents.networks import value_network
from tf_agents.networks import value_rnn_network
from tf_agents.policies import greedy_policy

from tf_agents.agents.ppo import ppo_agent
from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.trajectories import time_step as ts

from bark_ml.library_wrappers.lib_tf_agents.agents.tfa_agent import BehaviorTFAAgent
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.commons.py_spaces import BoundedContinuous


class BehaviorPPOAgent(BehaviorTFAAgent):
  def __init__(self,
               environment=None,
               params=None,
               observer=None):
    self._ppo_params = params["ML"]["BehaviorPPOAgent"]
    BehaviorTFAAgent.__init__(
      self,
      environment=environment,
      params=params,
      observer=observer)
    self._replay_buffer = self.GetReplayBuffer()
    self._collect_policy = self.GetCollectionPolicy()
    self._eval_policy = self.GetEvalPolicy()

  def GetAgent(self, env, params):
    actor_net = actor_distribution_network.ActorDistributionNetwork(
      env.observation_spec(),
      env.action_spec(),
      fc_layer_params=tuple(self._ppo_params[
        "ActorFcLayerParams", "", [512, 256, 256]]))
    value_net = value_network.ValueNetwork(
      env.observation_spec(),
      fc_layer_params=tuple(self._ppo_params[
        "CriticFcLayerParams", "", [512, 256, 256]]))

    tf_agent = ppo_agent.PPOAgent(
      env.time_step_spec(),
      env.action_spec(),
      actor_net=actor_net,
      value_net=value_net,
      normalize_observations=self._ppo_params[
        "NormalizeObservations", "", False],
      normalize_rewards=self._ppo_params["NormalizeRewards", "", False],
      optimizer=tf.compat.v1.train.AdamOptimizer(
        learning_rate=self._ppo_params["LearningRate", "", 3e-4]),
      train_step_counter=self._ckpt.step,
      num_epochs=self._ppo_params["NumEpochs", "", 25],
      name=self._ppo_params["AgentName", "", "ppo_agent"],
      debug_summaries=self._ppo_params["DebugSummaries", "", False])
    tf_agent.initialize()
    return tf_agent

  def GetReplayBuffer(self):
    return tf_uniform_replay_buffer.TFUniformReplayBuffer(
      data_spec=self._agent.collect_data_spec,
      batch_size=self._ppo_params["NumParallelEnvironments", "", 1],
      max_length=self._ppo_params["ReplayBufferCapacity", "", 1000])

  def GetCollectionPolicy(self):
    return self._agent.collect_policy

  def GetEvalPolicy(self):
    return self._agent.policy

  def Reset(self):
    pass

  @property
  def collect_policy(self):
    return self._collect_policy

  @property
  def eval_policy(self):
    return self._eval_policy


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import os.path

from bark.runtime.scenario import Scenario
from bark.runtime.scenario.scenario_generation import ScenarioGeneration
from bark.runtime.scenario.interaction_dataset_processing.interaction_dataset_reader import InteractionDatasetReader
from bark.runtime.commons.model_json_conversion import ModelJsonConversion
from bark.runtime.commons import ParameterServer
# PyBind imports
from bark.core.world.map import *
from bark.core.models.dynamic import *
from bark.core.models.execution import *
from bark.runtime.scenario.interaction_dataset_processing.scenario_track_info import ScenarioTrackInfo
from bark.runtime.scenario.interaction_dataset_processing.agent_track_info import AgentTrackInfo
import os.path


class InteractionDatasetScenarioGeneration(ScenarioGeneration):

    def __init__(self, params=None, num_scenarios=None, random_seed=None):
        super().__init__(params, num_scenarios, random_seed)

    def initialize_params(self, params):
        super().initialize_params(params)
        params_temp = \
            self._params["Scenario"]["Generation"]["InteractionDatasetScenarioGeneration"]
        self._map_file_name = params_temp["MapFilename",
                                          "Path to the open drive map",
                                          os.path.expanduser('~') + "/bark-simulator/interaction_dataset_fortiss_internal/DR_DEU_Merging_MT/map/DR_DEU_Merging_MT_v01_shifted.xodr"]
        self._track_file_name = params_temp["TrackFilename",
                                            "Path to track file (csv)",
                                            os.path.expanduser('~') + "/bark-simulator/interaction_dataset_fortiss_internal/DR_DEU_Merging_MT/tracks/vehicle_tracks_013.csv"]
        self._track_ids = params_temp["TrackIds",
                                      "IDs of the vehicle tracks to import.",
                                      [1]]
        self._xy_offset = params_temp["XYOffset",
                                      "offset in x and y direction.", [0, 0]]
        self._start_time = params_temp["StartTs",
                                       "Timestamp when to start the scenario (ms)", 0]
        self._end_time = params_temp["EndTs",
                                     "Timestamp when to end the scenario (ms)", None]
        self._ego_track_id = params_temp["EgoTrackId", "TrackID of ego", -1]
        self._behavior_models = params_temp["BehaviorModel",
                                            "Overwrite static trajectory with prediction model", {}]
        self._use_rectangle_shape = params_temp["RectangleShape",
                                                "Use Rectangle vehicle shape", True]
        
        self._interaction_ds_reader = InteractionDatasetReader()

    # TODO: remove code duplication with configurable scenario generation
    def create_scenarios(self, params, num_scenarios):
        """
            see baseclass
        """

        if num_scenarios != 1:
            raise ValueError("Replay supports only 1 scenario")

        scenario_track_info = self.__fill_scenario_track_info__()
        scenario_list = [self.__create_single_scenario__(scenario_track_info)]

        return scenario_list

    def __fill_scenario_track_info__(self):
        if self._ego_track_id == -1:
            raise ValueError("No ego id has been defined")

        ego_track_info = AgentTrackInfo(filename=self._track_file_name, track_id=self._ego_track_id,
                                        start_time=self._start_time, end_time=self._end_time)
        scenario_track_info = ScenarioTrackInfo(track_filename=self._track_file_name, ego_track_info=ego_track_info, xy_offset=self._xy_offset)
        for track_id in self._track_ids:
            new_agent = AgentTrackInfo(filename=self._track_file_name, track_id=track_id,
                                       start_time=self._start_time, end_time=self._end_time)
            scenario_track_info.AddTrackInfoOtherAgent(new_agent)

        return scenario_track_info

    def __fill_agent_params__(self):
        return self._params

    def __create_single_scenario__(self, scenario_track_info):
        scenario_track_info.TimeSanityCheck()

        scenario = Scenario(map_file_name=self._map_file_name,
                            json_params=self._params.ConvertToDict())

        world = scenario.GetWorldState()
        track_params = ParameterServer()
        track_params["execution_model"] = 'ExecutionModelInterpolate'
        track_params["dynamic_model"] = 'SingleTrackModel'
        track_params["map_interface"] = world.map

        all_track_ids = list(scenario_track_info.GetOtherTrackInfos().keys())
        # also add ego id
        all_track_ids.append(
            scenario_track_info.GetEgoTrackInfo().GetTrackId())

        agent_list = []
        model_converter = ModelJsonConversion()
        for track_id in all_track_ids:
            if str(track_id) in self._behavior_models:
                behavior_params = self.__fill_agent_params__()
                behavior_model_name = self._behavior_models[str(track_id)]
                track_params["behavior_model"] = model_converter.convert_model(
                    behavior_model_name, behavior_params)
            else:
                track_params["behavior_model"] = None

            agent = self._interaction_ds_reader.AgentFromTrackfile(
                track_params, self._params, scenario_track_info, track_id, goal_def=None)
            # agent_params.Save("/tmp/agent_params_{}.json".format(track_id))
            agent_list.append(agent)

        scenario._agent_list = agent_list  # must contain all agents!
        scenario._eval_agent_ids = [
            scenario_track_info.GetEgoTrackInfo().GetTrackId()]

        return scenario



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
from bark.runtime.runtime import Runtime
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark.runtime.commons.parameters import ParameterServer
from bark.core.models.behavior import BehaviorModel, BehaviorDynamicModel


class SingleAgentRuntime(Runtime):
  def __init__(self,
               blueprint=None,
               ml_behavior=None,
               observer=None,
               evaluator=None,
               step_time=None,
               viewer=None,
               scenario_generator=None,
               render=False):
    
    if blueprint is not None:
      self._scenario_generator = blueprint._scenario_generation
      self._viewer = blueprint._viewer
      self._ml_behavior = blueprint._ml_behavior
      self._step_time = blueprint._dt
      self._evaluator = blueprint._evaluator
      self._observer = blueprint._observer
    Runtime.__init__(self,
                     step_time=step_time or self._step_time,
                     viewer=viewer or self._viewer,
                     scenario_generator=scenario_generator or self._scenario_generator,
                     render=render)
    self._ml_behavior = ml_behavior or self._ml_behavior
    self._observer = observer or self._observer
    self._evaluator = evaluator or self._evaluator

  def reset(self, scenario=None):
    """Resets the runtime and its objects
    """
    super().reset(scenario=scenario)
    assert len(self._scenario._eval_agent_ids) == 1, \
      "This runtime only supports a single agent!"
    eval_id = self._scenario._eval_agent_ids[0]
    self._world.UpdateAgentRTree()
    self._world = self._observer.Reset(self._world)
    self._world = self._evaluator.Reset(self._world)
    self._world.agents[eval_id].behavior_model = self._ml_behavior
        
    # observe
    observed_world = self._world.Observe([eval_id])[0]
    return self._observer.Observe(observed_world)

  def step(self, action):
    # render
    if self._render:
      self.render()
      
    # set actions
    eval_id = self._scenario._eval_agent_ids[0]
    if eval_id in self._world.agents:
      self._world.agents[eval_id].behavior_model.ActionToBehavior(action)
    
    # step and observe
    self._world.Step(self._step_time)
    observed_world = self._world.Observe([eval_id])
    
    if len(observed_world) > 0:
      observed_world = observed_world[0]
    else:
      raise Exception('No world instance available.')

    # observe and evaluate
    observed_next_state = self._observer.Observe(observed_world)
    reward, done, info = self._evaluator.Evaluate(
      observed_world=observed_world,
      action=action)

    return observed_next_state, reward, done, info

  @property
  def action_space(self):
    """Action space of the agent
    """
    return self._ml_behavior.action_space

  @property
  def observation_space(self):
    """Observation space of the agent
    """
    return self._observer.observation_space

  @property
  def ml_behavior(self):
    return self._ml_behavior

  @ml_behavior.setter
  def ml_behavior(self, ml_behavior):
    self._ml_behavior = ml_behavior
    



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from .agent_state_geometry_config_readers import *
from .behavior_model_config_readers import *
from .controlled_agents_config_readers import *
from .dynamic_model_config_readers import *
from .execution_model_config_readers import *
from .goal_definition_config_readers import *


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from lxml import etree
import pprint
import logging
from bark.core.world.opendrive import *
from bark.core.geometry import Point2d
logger = logging.getLogger()


class XodrParser(object):
    def __init__(self, file_name, **kwargs):
        self._xodr = self.load_xodr(file_name)
        self._s_inc_straight_line = kwargs.pop("s_inc_straight_line", None)
        self._s_inc_curves = kwargs.pop("s_inc_curves", 0.2)
        self._python_map = {}
        
        self.parse_xml(self._xodr)
        self.map = OpenDriveMap()
        self.convert_to_map(self._python_map)

    def load_file(self, file_name):
        return open(file_name, 'r')

    def load_xodr(self, file_name):
        """Function that loads XODR and returns root node of

    Arguments:
      file_name {[string]} -- [File name of XODR]

    Raises:
      TypeError -- [If the file could be parsed]

    Returns:
      [etree] -- [Returns the root node of the etree]
    """
        with self.load_file(file_name) as xodr_file:
            parsed_xodr_file = etree.parse(xodr_file).getroot()
            if not etree.iselement(parsed_xodr_file):
                raise TypeError("Not a valid XODR format.")
            return parsed_xodr_file

    def parse_lane_link(self, lane_link):
        new_link = {}
        if lane_link.find("predecessor") is not None:
            new_link["predecessor"] = lane_link.find("predecessor").get("id")
        if lane_link.find("successor") is not None:
            new_link["successor"] = lane_link.find("successor").get("id")
        return new_link

    def parse_lane_road_mark(self, road_mark):
        new_road_mark = {}
        if str(road_mark.get("type")) in ["solid", "broken"]:
            new_road_mark["s_offset"] = road_mark.get("sOffset")
            new_road_mark["type"] = XodrRoadMarkType.__members__[
                str(road_mark.get("type"))]  # assign enum type # road_mark.get("type")
            new_road_mark["weight"] = road_mark.get("weight")
            new_road_mark["color"] = XodrRoadMarkColor.__members__[
                str(road_mark.get("color"))]  # assign enum type # road_mark.get("type")
            new_road_mark["width"] = road_mark.get("width")
        return new_road_mark

    def parse_lane_width(self, lane_width):
        new_lane_width = {}
        new_lane_width["s_offset"] = lane_width.get("sOffset")
        new_lane_width["a"] = lane_width.get("a")
        new_lane_width["b"] = lane_width.get("b")
        new_lane_width["c"] = lane_width.get("c")
        new_lane_width["d"] = lane_width.get("d")
        return new_lane_width

    def zero_lane_width(self):
        # TODO: do we really need this?
        new_lane_width = {}
        new_lane_width["s_offset"] = 0.0
        new_lane_width["a"] = 0.0
        new_lane_width["b"] = 0.0
        new_lane_width["c"] = 0.0
        new_lane_width["d"] = 0.0
        return new_lane_width

    def parse_lane_widths_from_lane(self, lane, id):
        lane_width_list = []
        lane_widths = lane.findall("width")

        if len(lane_widths) > 0:
            for lane_width in lane_widths:
                lane_width = lane_width
                lane_width_list.append(self.parse_lane_width(lane_width))
        else:
            if int(id) == 0:
                lane_width_list.append(self.zero_lane_width())

        return lane_width_list

    def parse_lanes_from_lane_sections(self, lanes, lane_section):
        # previous or next polynoamial
        lane_dict = {}
        for lane in lanes:
            lane_dict[int(lane.get("id"))] = lane

        for id, lane in lane_dict.items():
            new_lane = {}
            new_lane["id"] = id
            # every type we cannot read is read in as sidewalk
            new_lane["type"] = XodrLaneType.__members__[str(lane.get("type"))] if str(lane.get("type")) in [
                "driving", "border", "sidewalk"] else XodrLaneType.__members__["sidewalk"]  # assign enum type

            if lane.find("userData") is not None:
                if lane.find("userData").find("vectorLane") is not None:
                    vector_lane = lane.find("userData").find("vectorLane")
                    if str(vector_lane.get("travelDir")) in ["forward", "backward"]:
                        new_lane["driving_direction"] = XodrDrivingDirection.__members__[
                            str(vector_lane.get("travelDir"))]

            if "driving_direction" not in new_lane:
                if int(lane.get("id")) < 0:
                    new_lane["driving_direction"] = XodrDrivingDirection.forward
                else:
                    new_lane["driving_direction"] = XodrDrivingDirection.backward

            new_lane["level"] = lane.get("level")
            if lane.find("link") is not None:
                new_lane["link"] = self.parse_lane_link(lane.find("link"))
            if lane.find("roadMark") is not None:
                road_mark = self.parse_lane_road_mark(lane.find("roadMark"))
                if road_mark:  # if dict is not empty
                    new_lane["road_mark"] = road_mark

            new_lane["width"] = self.parse_lane_widths_from_lane(lane, id)

            lane_section["lanes"].append(new_lane)
        return lane_section

    def parse_offset(self, header):
        # in compiliance with 5.2.2 of specification
        offset = {}
        offset["x"] = float(header.find("offset").get("x"))
        offset["y"] = float(header.find("offset").get("y"))
        offset["z"] = float(header.find("offset").get("z"))
        offset["hdg"] = float(header.find("offset").get("hdg"))
        return offset

    def parse_header(self, header):
        new_header = {}
        new_header["north"] = header.get("north")
        new_header["south"] = header.get("south")
        new_header["east"] = header.get("east")
        new_header["west"] = header.get("west")
        if header.find("offset") is not None:
            new_header["offset"] = self.parse_offset(header)
        self._python_map["header"] = new_header

    def parse_lane_sections_from_road(self, lane_sections, road, lane_offset):
        road["lane_offset"] = lane_offset
        road["lane_sections"] = []
        for lane_section in lane_sections:
            new_lane_section = {}
            new_lane_section["s"] = lane_section.get("s")
            lane_keys = ['left', 'center', 'right']
            new_lane_section["lanes"] = []
            for key in lane_keys:
                if lane_section.find(key) is not None:
                    for lane_group in lane_section.findall(key):
                        new_lane_section = self.parse_lanes_from_lane_sections(
                            lane_group.findall('lane'), new_lane_section)
            road["lane_sections"].append(new_lane_section)
        return road

    def parse_plan_view(self, plan_view, road):
        geometries = plan_view.findall("geometry")
        new_plan_view = {}
        new_plan_view["geometries"] = []
        for geometry in geometries:
            new_geometry = {}
            new_geometry["s"] = geometry.get("s")
            new_geometry["x"] = geometry.get("x")
            new_geometry["y"] = geometry.get("y")
            new_geometry["hdg"] = geometry.get("hdg")
            new_geometry["length"] = geometry.get("length")
            if geometry.find("line") is not None:
                new_line = {"type": "line"}
                new_geometry["geometry"] = new_line
            if geometry.find("spiral") is not None:
                new_spiral = {"type": "spiral"}
                new_spiral["curv_start"] = geometry.find("spiral").get(
                    "curvStart")
                new_spiral["curv_end"] = geometry.find("spiral").get("curvEnd")
                new_geometry["geometry"] = new_spiral
            if geometry.find("arc") is not None:
                new_arc = {"type": "arc"}
                new_arc["curvature"] = geometry.find("arc").get("curvature")
                new_geometry["geometry"] = new_arc
            new_plan_view["geometries"].append(new_geometry)
        road["plan_view"] = new_plan_view
        return road

    def parse_road_link_pre_suc(self, element):
        new_pre_suc = {}
        if element.get("elementType") is not None:
            new_pre_suc["element_type"] = element.get("elementType")
        if element.get("elementId") is not None:
            new_pre_suc["element_id"] = element.get("elementId")
        if element.get("contactPoint") is not None:
            new_pre_suc["contact_point"] = element.get("contactPoint")
        return new_pre_suc

    def parse_road_link(self, road_link):
        new_road_link = {}
        if road_link.find("predecessor") is not None:
            new_road_link["predecessor"] = self.parse_road_link_pre_suc(
                road_link.find("predecessor"))
        if road_link.find("successor") is not None:
            new_road_link["successor"] = self.parse_road_link_pre_suc(
                road_link.find("successor"))
        return new_road_link

    def parse_road(self, road):
        new_road = {}
        new_road["id"] = road.get("id")
        new_road["length"] = road.get("length")
        new_road["junction"] = road.get("junction")
        new_road["name"] = road.get("name")
        lanes = road.find("lanes")
        lane_offset = self.parse_lane_offset(lanes.find("laneOffset"))
        lane_sections = lanes.findall("laneSection")
        new_road = self.parse_lane_sections_from_road(
            lane_sections, new_road, lane_offset)
        new_road = self.parse_plan_view(road.find("planView"), new_road)
        if road.find("link") is not None:
            new_road["link"] = self.parse_road_link(road.find("link"))
        self._python_map["roads"].append(new_road)

    def parse_junction_links(self, connection):
        new_links = []
        for link in connection.findall("laneLink"):
            new_link = {}
            new_link["from"] = link.get("from")
            new_link["to"] = link.get("to")
            new_links.append(new_link)
        return new_links

    def parse_junctions(self, junctions):
        for junction in junctions:
            new_junction = {}
            new_junction["id"] = junction.get("id")
            new_junction["connections"] = []
            for connection in junction.findall("connection"):
                new_connection = {}
                new_connection["id"] = connection.get("id")
                new_connection["incoming_road"] = connection.get(
                    "incomingRoad")
                new_connection["connecting_road"] = connection.get(
                    "connectingRoad")
                new_connection["contact_point"] = connection.get(
                    "contactPoint")
                new_connection["lane_links"] = self.parse_junction_links(
                    connection)
                new_junction["connections"].append(new_connection)
            self._python_map["junctions"].append(new_junction)

    def parse_xml(self, xodr_obj):
        """Imports the XODR file to python

    Arguments:
      xodr_obj {etree} -- containing map information

    Returns:
      python dict -- containing all neccessary map information
    """
        self.parse_header(xodr_obj.find("header"))
        self._python_map["roads"] = []
        self._python_map["junctions"] = []
        for road in xodr_obj.findall("road"):
            self.parse_road(road)
        self.parse_junctions(xodr_obj.findall("junction"))

    def print_python_map(self):
        pp = pprint.PrettyPrinter(indent=2)
        pp.pprint(self._python_map)

    def create_cpp_plan_view(self, plan_view, header):

        new_plan_view = PlanView()
        # create plan view..
        for geo in plan_view["geometries"]:
            start_p = Point2d(float(geo["x"]), float(geo["y"]))
            if geo["geometry"]["type"] == "line":
                if self._s_inc_straight_line is None:
                    s_inc_straight_line = float(geo["length"])
                elif isinstance(self._s_inc_straight_line, float):
                    s_inc_straight_line = self._s_inc_straight_line
                else:
                    raise TypeError("s_inc_straight_line not specified")
                new_plan_view.AddLine(start_p, float(geo["hdg"]),
                                      float(geo["length"]), s_inc_straight_line)

            if geo["geometry"]["type"] == "arc":
                new_plan_view.AddArc(start_p, float(geo["hdg"]), float(
                    geo["length"]), float(geo["geometry"]["curvature"]), self._s_inc_curves)

            if geo["geometry"]["type"] == "spiral":
                new_plan_view.AddSpiral(start_p, float(geo["hdg"]), float(geo["length"]), float(
                    geo["geometry"]["curv_start"]), float(geo["geometry"]["curv_end"]), self._s_inc_curves)

        # now use header/ offset to modify plan view
        if "offset" in header:
            off_x = header["offset"]["x"]
            off_y = header["offset"]["y"]
            off_hdg = header["offset"]["hdg"]
            logger.info("Transforming PlanView with given offset",
                        header["offset"])
            new_plan_view.ApplyOffsetTransform(off_x, off_y, off_hdg)

        return new_plan_view

    def create_cpp_road_link(self, link):
        new_link = XodrRoadLink()
        try:
            new_pre_info = XodrRoadLinkInfo()
            new_pre_info.id = int(link["predecessor"]["element_id"])
            new_pre_info.type = link["predecessor"]["element_type"]
            new_link.predecessor = new_pre_info
        except:
            pass

        try:
            new_suc_info = XodrRoadLinkInfo()
            new_suc_info.id = int(link["successor"]["element_id"])
            new_suc_info.type = link["successor"]["element_type"]
            new_link.successor = new_suc_info
        except:
            pass
        return new_link

    def create_cpp_road(self, road, header):
        new_road = XodrRoad()
        new_road.id = int(road["id"])
        new_road.name = road["name"]
        new_road.plan_view = self.create_cpp_plan_view(
            road["plan_view"], header)
        new_road = self.create_cpp_lane_section(new_road, road)

        try:
            new_road.link = self.create_cpp_road_link(road["link"])
        except:
            pass

        return new_road

    def create_lane_link(self, link):
        new_link = XodrLaneLink()

        if link is not None:
            try:
                new_link.from_position = int(link["predecessor"])
            except:
                pass
            try:
                new_link.to_position = int(link["successor"])
            except:
                pass
        else:
            logger.debug("No XodrLaneLink")

        return new_link

    def create_cpp_lane(self, new_lane_section, new_road, lane, s_end, reference_line, lane_offset):

        try:
            new_lane = XodrLane(int(lane["id"]))
            for idx_w, lw in enumerate(lane["width"]):

                a = float(lane["width"][idx_w]["a"]) + lane_offset["a"]
                b = float(lane["width"][idx_w]["b"]) + lane_offset["b"]
                c = float(lane["width"][idx_w]["c"]) + lane_offset["c"]
                d = float(lane["width"][idx_w]["d"]) + lane_offset["d"]
                offset = XodrLaneOffset(a, b, c, d)

                s_start_temp = float(
                    lane["width"][idx_w]["s_offset"]) + lane_offset["s"]

                if idx_w < len(lane["width"]) - 1:
                    s_end_temp = float(lane["width"][idx_w+1]["s_offset"])
                else:
                    # last or only lane width element
                    s_end_temp = s_end

                lane_width = XodrLaneWidth(s_start_temp, s_end_temp, offset)

                # TODO (@hart): make sampling flexible
                succ = new_lane.append(reference_line, lane_width, 0.2)

            new_lane.lane_type = lane["type"]
            new_lane.driving_direction = lane["driving_direction"]

            if "link" in lane:
                new_lane.link = self.create_lane_link(lane["link"])
            else:
                new_lane.link = self.create_lane_link(None)

            # not every lane contains a road-mark
            if ("road_mark" in lane):
                rm = XodrRoadMark()
                rm.type = lane['road_mark']['type']
                rm.color = lane['road_mark']['color']
                rm.width = float(lane['road_mark']['width'])
                new_lane.road_mark = rm

            new_lane_section.AddLane(new_lane)

        except:
            raise ValueError("Something went wrong with creating the lane.")
        return new_lane_section

    def sortedIndices(self, lst):
        return sorted(range(len(lst)), key=lambda x: (abs(lst[x])))

    def create_cpp_lane_section(self, new_road, road):

        # In xodr the <lane> tag can have a <laneOffset> tag, that shifts all lanes by this 
        # constant offset. We here shift the inner lanes explicitly by the offset and the outer lanes
        #  (|id| > 1) are then shifted implicitly.
        lane_offset = road["lane_offset"]
        no_lane_offset = {"s": 0.0, "a": 0.0, "b": 0.0, "c": 0.0, "d": 0.0}
        for lane_section in road["lane_sections"]:
            new_lane_section = XodrLaneSection(float(lane_section["s"]))
            # sort lanes
            # for idx_iterator in range(len(lane_section["lanes"])):
            #    if lane_section["lanes"][idx_iterator]['id'] in list_id_read_in_lanes:

            unordered_id_list = [l['id'] for l in lane_section["lanes"]]
            indices = self.sortedIndices(unordered_id_list)

            # for idx, lane in enumerate(lane_section["lanes"]):
            for idx_iterator in indices:
                lane = lane_section["lanes"][idx_iterator]
                if lane['id'] == 0:
                    # plan view
                    new_lane_section = self.create_cpp_lane(new_lane_section, new_road, lane, float(
                        road["length"]), new_road.plan_view.GetReferenceLine(), no_lane_offset)
                elif lane['id'] == -1 or lane['id'] == 1:
                    # use plan view for offset calculation
                    new_lane_section = self.create_cpp_lane(new_lane_section, new_road, lane, float(
                        road["length"]), new_road.plan_view.GetReferenceLine(), lane_offset)
                else:
                    # use previous line for offset calculation
                    #temp_lanes = new_lane_section.GetLanes()

                    if lane['id'] > 0:
                        previous_line = new_lane_section.GetLaneByPosition(
                            lane['id']-1).line
                        new_lane_section = self.create_cpp_lane(
                            new_lane_section, new_road, lane, previous_line.Length(), previous_line, no_lane_offset)
                    elif lane['id'] < 0:
                        previous_line = new_lane_section.GetLaneByPosition(
                            lane['id']+1).line
                        new_lane_section = self.create_cpp_lane(
                            new_lane_section, new_road, lane, previous_line.Length(), previous_line, no_lane_offset)
                    else:
                        logger.info("Calculating previous lane did not work.")

            new_road.AddLaneSection(new_lane_section)
        return new_road

    def GetMap_element(self, key, id):
        for x in self._python_map[key]:
            if x["id"] == id:
                return x

    def create_cpp_junction(self, junction):
        new_junction = Junction("", int(junction["id"]))
        for connection in junction["connections"]:
            new_connection = Connection()
            new_connection.incoming_road = int(connection["incoming_road"])
            new_connection.connecting_road = int(connection["connecting_road"])
            new_connection.id = int(connection["id"])
            # TODO(hart): contact point
            for lane_link in connection["lane_links"]:
                new_lane_link = XodrLaneLink()
                new_lane_link.from_position = int(lane_link["from"])
                new_lane_link.to_position = int(lane_link["to"])
                new_connection.AddLaneLink(new_lane_link)
            new_junction.AddConnection(new_connection)
        return new_junction

    def convert_to_map(self, python_map):
        """Loops through python_map and converts it to a cpp map

    Arguments:
      python_map {dict} -- containing all the map info

    Returns:
      CPP Map -- Map for usage with CPP
    """
        for road in self._python_map["roads"]:
            new_road = self.create_cpp_road(road, self._python_map["header"])
            self.map.AddRoad(new_road)

        for junction in self._python_map["junctions"]:
            new_junction = self.create_cpp_junction(junction)
            self.map.AddJunction(new_junction)

    def parse_lane_offset(self, lane_offset):
        if lane_offset is not None:
            s = float(lane_offset.get("s"))
            a = float(lane_offset.get("a"))
            b = float(lane_offset.get("b"))
            c = float(lane_offset.get("c"))
            d = float(lane_offset.get("d"))
        else:
            s = 0.0
            a = 0.0
            b = 0.0
            c = 0.0
            d = 0.0
        return {"s": s, "a": a, "b": b, "c": c, "d": d}




# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import math
from abc import ABC, abstractmethod


class StateObserver(ABC):
  def __init__(self,
               params):
    self._params = params
    self._velocity_range = \
      self._params["ML"]["StateObserver"]["VelocityRange",
      "Boundaries for min and max velocity for normalization",
      [0, 100]]
    self._theta_range = \
      self._params["ML"]["StateObserver"]["ThetaRange",
      "Boundaries for min and max theta for normalization",
      [-2*math.pi, 2*math.pi]]
    self._normalization_enabled = \
      self._params["ML"]["StateObserver"]["NormalizationEnabled",
      "Whether normalization should be performed",
      True]
    self._max_num_vehicles = \
      self._params["ML"]["StateObserver"]["MaxNumAgents",
      "The concatenation state size is the ego agent plus max num other agents",
      2]
    self._world_x_range = [-10000, 10000]
    self._world_y_range = [-10000, 10000]

  @abstractmethod
  def Observe(self, observed_world):
    """Observes the world
    
    Arguments:
        world {bark.ObservedWorld} -- observed BARK world
        agents_to_observe {list(int)} -- ids of agents to observe
    
    Returns:
        np.array -- concatenated state array
    """
    pass
  
  def _select_state_by_index(self, state):
    """selects a subset of an array using the state definition
    
    Arguments:
        state {np.array} -- full state space
    
    Returns:
        np.array -- reduced state space
    """
    return state[self._state_definition]

  def Reset(self, world):
    bb = world.bounding_box
    self._world_x_range = [bb[0].x(), bb[1].x()]
    self._world_y_range = [bb[0].y(), bb[1].y()]
    return world

  @property
  def observation_space(self):
    pass


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.


from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces import ConfigReaderGoalDefinitions

from bark.core.world.goal_definition import GoalDefinitionStateLimitsFrenet
from bark.runtime.commons.parameters import ParameterServer

from bark.core.geometry import *
import logging

# helper class to support various types of goals easily generated for the 
# goal config readers
class GoalGenerator:
  @staticmethod
  def get_goal_definition(config_param_object, goal_definition, road_corridor, lane_position, enforce_goal):
    if enforce_goal:
      return eval("GoalGenerator.{}(config_param_object, road_corridor, lane_position)".format(goal_definition))
    else:
      try:
        return eval("GoalGenerator.{}(config_param_object, road_corridor, lane_position)".format(goal_definition))
      except:
        return eval("GoalGenerator.EndOfLane(config_param_object, road_corridor, lane_position)")

  @staticmethod
  def EndOfLane(config_param_object, road_corridor, lane_position):
      lane_corridor = road_corridor.lane_corridors[lane_position]
      return GoalGenerator.fill_goal_def_state_lim_frenet(config_param_object, lane_corridor.center_line)

  @staticmethod
  def LaneChangeRight(config_param_object, road_corridor, lane_position):
    lane_position_right = lane_position+1
    if lane_position_right <  len(road_corridor.lane_corridors):
      lane_corridor_right = road_corridor.lane_corridors[lane_position_right]
      return GoalGenerator.fill_goal_def_state_lim_frenet(config_param_object, lane_corridor_right.center_line)
    else:
      raise ValueError("No lane change right possible with this \
               road corridor at lane position {}".format(lane_position))

  @staticmethod
  def LaneChangeLeft(config_param_object, road_corridor, lane_position):
    lane_position_left = lane_position-1
    if lane_position_left >= 0:
      lane_position_left = road_corridor.lane_corridors[lane_position_left]
      return GoalGenerator.fill_goal_def_state_lim_frenet(config_param_object, lane_position_left.center_line)
    else:
      raise ValueError("No lane change left possible with this \
               road corridor at lane position {}".format(lane_position))

  @staticmethod
  def fill_goal_def_state_lim_frenet(config_param_object, line_string):
    lateral_max_dist = config_param_object["MaxLateralDist", "Pair with values between 0,1: \
          Lateral maximum distance allowed to both sides of center line, normalized by lanewidth", (0.1, 0.1)]
    long_range = config_param_object["LongitudinalRange" , "Pair with values between 0,1: \
          Goal is within this longitudinal part of center line, normalized by lanewidth", (0, 1.0)]
    max_orientation_diff = config_param_object["MaxOrientationDifference" , "Pair with values between 0,pi: \
          Orientation must be within orientation limits around tangent angle of center line", (0.08, 0.08)]
    velocity_range = config_param_object["VelocityRange" , "Pair velocity values specifying allowed range", (10, 20)]

    length = line_string.Length()
    goal_line_string = GetLineFromSInterval(line_string, long_range[0]*length, long_range[1]*length)
    goal_definition = GoalDefinitionStateLimitsFrenet(goal_line_string, lateral_max_dist, max_orientation_diff, velocity_range)
    return goal_definition


class FixedGoalTypes(ConfigReaderGoalDefinitions):
  def create_from_config(self, config_param_object, road_corridor, agent_states, controlled_agent_ids, **kwargs):
    self._controlled_agents_goal_type = config_param_object["GoalTypeControlled", "Specifies type of goals \
                          for controlled agents (EndOfLane, LaneChangeLeft, LaneChangeRight)", "EndOfLane"]
    self._enforce_controlled_goal = config_param_object["EnforceControlledGoal", "If true exception is raised if goal not available", True]
    
    self._other_agents_goal_type = config_param_object["GoalTypeOthers", "Specifies type of goals \
                          for other agents (EndOfLane, LaneChangeLeft, LaneChangeRight)", "EndOfLane"]
    self._enforce_others_goal = config_param_object["EnforceOthersGoal", "If true exception is raised if goal not available", True]

    goal_definitions = []
    agent_lane_positions = kwargs.pop("agent_lane_positions")
    for idx, _ in enumerate(agent_states):
      lane_position = agent_lane_positions[idx]
      if isinstance(lane_position, list):
        lane_position = lane_position[0]
      if controlled_agent_ids[idx]:
          goal_definition = GoalGenerator.get_goal_definition(
            config_param_object, self._controlled_agents_goal_type, road_corridor, lane_position, self._enforce_controlled_goal)
      else:
          goal_definition = GoalGenerator.get_goal_definition(
          config_param_object, self._other_agents_goal_type, road_corridor, lane_position, self._enforce_others_goal)
      goal_definitions.append(goal_definition)
    return goal_definitions, {}, config_param_object







# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import tempfile
import os
import shutil
import uuid
import subprocess
import logging

from bark.core.geometry import *
from bark.core.viewer import *
from bark.core.models.dynamic import *
from bark.runtime.viewer.viewer import BaseViewer


class VideoRenderer(BaseViewer):
    def __init__(self, renderer, world_step_time, render_intermediate_steps=None , params=None, video_name=None, fig_path=None, clear_figures=True, **kwargs):
        super(VideoRenderer, self).__init__(params=params)
        self.renderer = renderer
        self.world_step_time = world_step_time
        self.render_intermediate_steps = render_intermediate_steps

        video_dir_name = "vid_render_frames_{}".format(uuid.uuid4()) if video_name is None else video_name
        self.fig_path = fig_path if fig_path is not None else tempfile.gettempdir()
        self.video_frame_dir = os.path.join(self.fig_path, video_dir_name)
        if clear_figures and os.path.exists(self.video_frame_dir):
            print("video path already exists, clear existing one")
            self.reset()
        os.makedirs(self.video_frame_dir, exist_ok=True)

        self.frame_count = 0

    def drawWorld(self, world, eval_agent_ids=None, scenario_idx=None, debug_text=False):
        if self.render_intermediate_steps is None:
            self._renderWorld(world, eval_agent_ids, scenario_idx)
        else:
            world_time = world.time
            executed_world = world
            for _ in range(0,self.render_intermediate_steps):
                executed_world = executed_world.GetWorldAtTime(world_time)
                self._renderWorld(executed_world, eval_agent_ids, scenario_idx)
                world_time = world_time + self.world_step_time/self.render_intermediate_steps

    def drawText(self, **kwargs):
        self.renderer.drawText(**kwargs)

    def _renderWorld(self, world, eval_agent_ids=None, scenario_idx=None, debug_text=False):
        image_path = os.path.join(self.video_frame_dir, "{:03d}.png".format(self.frame_count))
        self.renderer.drawWorld(world=world, eval_agent_ids=eval_agent_ids,
                                filename=image_path, scenario_idx=scenario_idx, debug_text=debug_text)
        self.frame_count = self.frame_count + 1

    def reset(self):
        shutil.rmtree(os.path.abspath(self.video_frame_dir))

    def export_video(self, filename, remove_image_dir=True):
        if self.render_intermediate_steps is None:
            framerate = 1/self.world_step_time
        else:
            framerate =  1/(self.world_step_time/self.render_intermediate_steps)

        if os.path.isfile(filename):
            os.remove(filename)
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename))
        cmd = "ffmpeg -nostdin -y -framerate {} -i \'{}/%03d.png\' -vcodec h264 -force_key_frames 'expr:gte(t,n_forced*{})' -acodec pcm_s16le -s 1920x1080 -r 30 -b:v 36M -pix_fmt yuv420p -f mp4 \'{}.mp4\'".format(
            framerate, self.video_frame_dir, 1 / framerate, os.path.abspath(filename))
        retval = subprocess.call(cmd, shell=True)
        if retval:
          logging.error("Error during video export.")

        if remove_image_dir:
            shutil.rmtree(os.path.abspath(self.video_frame_dir))

    def clear(self):
      self.renderer.clear()






from bark_ml.observers.nearest_state_observer import NearestAgentsObserver
from bark_ml.observers.graph_observer import GraphObserver
from bark_ml.observers.graph_observer_v2 import GraphObserverV2
from bark_ml.observers.simple_observer import SimpleObserver


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
import tensorflow as tf
import logging

# BARK imports
from bark.core.models.behavior import BehaviorModel, BehaviorDynamicModel

# tfa
from tf_agents.networks import actor_distribution_network
from tf_agents.networks import normal_projection_network
from tf_agents.agents.ddpg import critic_network
from tf_agents.policies import greedy_policy
from tf_agents.environments import tf_py_environment
from tf_agents.agents.sac import sac_agent
from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.utils.common import Checkpointer
from tf_agents.trajectories import time_step as ts

# BARK-ML imports
from bark_ml.library_wrappers.lib_tf_agents.py_bark_environment import PyBARKEnvironment
from bark_ml.commons.py_spaces import BoundedContinuous
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML


class BehaviorTFAAgent(BehaviorModel):
  def __init__(self,
               environment=None,
               params=None,
               bark_behavior=None,
               observer=None):
    BehaviorModel.__init__(self, params)
    self._params = params
    self._observer = observer
    self._environment = environment
    self._wrapped_env = tf_py_environment.TFPyEnvironment(
      PyBARKEnvironment(self._environment))
    self._ckpt = tf.train.Checkpoint(step=tf.Variable(0, dtype=tf.int64))
    self._agent = self.GetAgent(self._wrapped_env, params)
    self._ckpt = tf.train.Checkpoint(step=tf.Variable(0, dtype=tf.int64),
                                     agent=self._agent)
    self._ckpt_manager = self.GetCheckpointer()
    self._logger = logging.getLogger()
    # NOTE: by default we do not want the action to be set externally
    #       as this enables the agents to be plug and played in BARK.
    self._set_action_externally = False
    self._bark_behavior_model = bark_behavior or BehaviorContinuousML(params)

  def Reset(self):
    pass
  
  @property
  def set_action_externally(self):
    return self._set_action_externally
  
  @set_action_externally.setter
  def set_action_externally(self, externally):
    # if externally:
    #   self._logger.info("Actions are now set externally.")
    self._set_action_externally = externally
  
  def GetCheckpointer(self):
    checkpointer = Checkpointer(
        self._params["ML"]["BehaviorTFAAgents"]["CheckpointPath", "", ""],
      global_step=self._ckpt.step,
      tf_agent=self._agent,
      max_to_keep=self._params["ML"]["BehaviorTFAAgents"][
        "NumCheckpointsToKeep", "", 3])
    checkpointer.initialize_or_restore()
    return checkpointer

  def Save(self):
    save_path = self._ckpt_manager.save(
      global_step=self._agent._train_step_counter)
    self._logger.info("Saved checkpoint for step {}.".format(
      int(self._agent._train_step_counter.numpy())))

  def Load(self):
    try:
      self._ckpt.restore(self._ckpt_manager.latest_checkpoint)
    except:
      return RuntimeError("Could not load agent.")
    if self._ckpt_manager.latest_checkpoint:
      self._logger.info("Restored agent from {}".format(
        self._ckpt_manager.latest_checkpoint))
    else:
      self._logger.info("Initializing agent from scratch.")

  def Act(self, state):
    # NOTE: greedy action
    action_step = self.eval_policy.action(
      ts.transition(state, reward=0.0, discount=1.0))
    return action_step.action.numpy()

  def ActionToBehavior(self, action):
    # NOTE: will either be set externally or internally
    self._action = action

  def Plan(self, dt, observed_world):
    # NOTE: if training is enabled the action is set externally
    if not self._set_action_externally:
      # NOTE: we need to store the observer differently
      observed_state = self._environment._observer.Observe(
        observed_world)
      self._action = self.Act(observed_state)
    # NOTE: BARK expects (m, 1) actions
    action = self._action
    if isinstance(self.action_space, BoundedContinuous):
      action = np.reshape(self._action, (-1, 1))
    # set action to be executed
    self._bark_behavior_model.ActionToBehavior(action)
    trajectory = self._bark_behavior_model.Plan(dt, observed_world)
    # NOTE: BARK requires models to have trajectories of the past
    BehaviorModel.SetLastTrajectory(self, trajectory)
    return trajectory

  @property
  def action_space(self):
    return self._bark_behavior_model.action_space

  def Clone(self):
    return self
  


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import logging
import tensorflow as tf
from enum import Enum
from graph_nets import modules
from graph_nets import utils_tf
from graph_nets import utils_np
from graph_nets.graphs import GraphsTuple
import sonnet as snt
import tensorflow_addons as tfa

# bark-ml
from bark.runtime.commons.parameters import ParameterServer
from bark_ml.observers.graph_observer import GraphObserver
from bark_ml.observers.graph_observer_v2 import GraphObserverV2
from bark_ml.library_wrappers.lib_tf_agents.networks.gnns.graph_network import GraphNetwork


def make_mlp(name, layer_size=150, embedding_size=80):
  return tf.keras.Sequential([
    tf.keras.layers.Dense(
      layer_size, activation='relu',
      bias_initializer=tf.constant_initializer(0.001),
      name=name+"_0"),
    tf.keras.layers.Dense(
      embedding_size, activation='relu',
      bias_initializer=tf.constant_initializer(0.001),
      name=name+"_1"),
    # tf.keras.layers.LayerNormalization()
  ])

class InteractionWrapper(GraphNetwork):
  """
  Implements a graph lib.
  """

  def __init__(self,
               params=ParameterServer(),
               name='InteractionNetwork',
               output_dtype=tf.float32):
    """
    Initializes a InteractionNetwork-Wrapper instance.

    Args:
    params: A `ParameterServer` instance containing the parameters
      to configure the GNN.
    name: Name of the instance.
    output_dtype: The dtype to which the GNN output is casted.
    """
    super(InteractionWrapper, self).__init__(
      params=params,
      name=name,
      output_dtype=output_dtype)
    self._num_message_passing_layers = params["ML"]["InteractionNetwork"][
      "NumMessagePassingLayers", "Number of message passing layers", 2]
    self._embedding_size = params["ML"]["InteractionNetwork"][
      "EmbeddingSize", "Embedding size of nodes", 20]
    
    self._node_mlps = []
    self._edge_mlps = []
    for i in range(0, self._num_message_passing_layers):
      self._node_mlps.append(
        make_mlp(name+"_node_"+str(i), embedding_size=self._embedding_size))
      self._edge_mlps.append(
        make_mlp(name+"_edge_"+str(i), embedding_size=self._embedding_size))

    self._latent_trace = None
    # initialize network & call func
    self._init_network(name)
    self._call_func = self._init_call_func
    
  def _init_network(self, name=None):
    self._graph_blocks = []
    for i in range(0, self._num_message_passing_layers):
      self._graph_blocks.append(
        modules.InteractionNetwork(
          edge_model_fn=lambda: self._edge_mlps[i],
          node_model_fn=lambda: self._node_mlps[i]))
  
  # @tf.function
  def _init_call_func(self, observations, training=False):
    """Graph nets implementation"""
    node_vals, edge_indices, node_to_graph, edge_vals = GraphObserver.graph(
      observations=observations, 
      graph_dims=self._graph_dims,
      dense=True)
    batch_size = tf.shape(observations)[0]
    node_counts = tf.unique_with_counts(node_to_graph)[2]
    edge_counts = tf.math.square(node_counts)
    
    input_graph = GraphsTuple(
      nodes=tf.cast(node_vals, tf.float32),
      edges=tf.cast(edge_vals, tf.float32),
      globals=tf.tile([[0.]], [batch_size, 1]),
      receivers=tf.cast(edge_indices[:, 1], tf.int32),
      senders=tf.cast(edge_indices[:, 0], tf.int32),
      n_node=node_counts,
      n_edge=edge_counts) 
    
    self._latent_trace = []
    latent = input_graph
    for gb in self._graph_blocks:
      latent = gb(latent)
      self._latent_trace.append(latent)
    node_values = tf.reshape(latent.nodes, [batch_size, -1, self._embedding_size])
    return node_values




from .scenario import  Scenario
from .scenario import SetMapfileDirectory
from .scenario import GetMapfileDirectory


__all__ = [
          "Scenario",
          "SetMapfileDirectory",
          "GetMapfileDirectory"
          ]



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import logging
import os

from bark.core.world.agent import Agent
from bark.core.models.behavior import BehaviorStaticTrajectory
from bark.core.models.dynamic import StateDefinition
from bark.core.world.goal_definition import GoalDefinition, GoalDefinitionPolygon
from bark.core.geometry import Point2d, Polygon2d, NormToPI
from bark.core.geometry.standard_shapes import *
from bark.runtime.commons.model_json_conversion import ModelJsonConversion
# Interaction dataset tools
from com_github_interaction_dataset_interaction_dataset.python.utils import dataset_reader
from com_github_interaction_dataset_interaction_dataset.python.utils import dict_utils
import numpy as np

from bark.runtime.scenario.interaction_dataset_processing.agent_track_info import AgentTrackInfo


def BarkStateFromMotionState(state, xy_offset, time_offset=0):
    bark_state = np.zeros(int(StateDefinition.MIN_STATE_SIZE))
    bark_state[int(StateDefinition.TIME_POSITION)] = (
        state.time_stamp_ms - time_offset) / 1000.0
    bark_state[int(StateDefinition.X_POSITION)] = state.x + xy_offset[0]
    bark_state[int(StateDefinition.Y_POSITION)] = state.y + xy_offset[1]
    orientation = NormToPI(state.psi_rad)
    if (orientation > np.pi or orientation < -np.pi):
        logging.error(
            "Orientation in Track file is ill-defined: {}".format(state.psi_rad))
    bark_state[int(StateDefinition.THETA_POSITION)] = orientation
    bark_state[int(StateDefinition.VEL_POSITION)] = pow(
        pow(state.vx, 2) + pow(state.vy, 2), 0.5)
    return bark_state.reshape((1, int(StateDefinition.MIN_STATE_SIZE)))


def TrajectoryFromTrack(track, xy_offset, start=0, end=None):
    states = list(dict_utils.get_item_iterator(track.motion_states))
    if end is None:
        end = states[-1][0]
    filtered_motion_states = list(
        filter(lambda s: start <= s[0] <= end, states))
    n = len(filtered_motion_states)
    traj = np.zeros((n, int(StateDefinition.MIN_STATE_SIZE)))
    for i, state in enumerate(filtered_motion_states):
        # TODO: rename start to be the scenario start time!
        traj[i, :] = BarkStateFromMotionState(
            state[1], xy_offset=xy_offset, time_offset=start)
    return traj


def ShapeFromTrack(track):
    r = ColRadiusFromTrack(track)
    wb = WheelbaseFromTrack(track)
    poly = GenerateCarRectangle(wb, r)
    return poly


def WheelbaseFromTrack(track):
    r = ColRadiusFromTrack(track)
    wb = track.length - 2*r
    return wb


def ColRadiusFromTrack(track):
    return track.width/2


def InitStateFromTrack(track, xy_offset, start):
    minimum_start = min(track.motion_states)
    if minimum_start > start:
        start = minimum_start
    state = track.motion_states[int(start)]
    bark_state = BarkStateFromMotionState(
        state, xy_offset=xy_offset, time_offset=state.time_stamp_ms)
    return bark_state.reshape((int(StateDefinition.MIN_STATE_SIZE), 1))


def GoalDefinitionFromTrack(track, end, xy_offset):
    goal_size = 12.0
    states = list(dict_utils.get_item_iterator(track.motion_states))
    # Goal position is spatial position of last state
    motion_state = states[-1][1]
    bark_state = BarkStateFromMotionState(motion_state, xy_offset=xy_offset)
    goal_polygon = Polygon2d(np.array([0.5 * goal_size, 0.5 * goal_size, 0.0]),
                             [Point2d(0.0, 0.0),
                              Point2d(goal_size, 0.0),
                              Point2d(goal_size, goal_size),
                              Point2d(0.0, goal_size),
                              Point2d(0.0, 0.0)])
    goal_point = Point2d(bark_state[0, int(StateDefinition.X_POSITION)] - 0.5 *
                         goal_size, bark_state[0, int(StateDefinition.Y_POSITION)] - 0.5 * goal_size)
    goal_polygon = goal_polygon.Translate(goal_point)
    goal_definition = GoalDefinitionPolygon(goal_polygon)
    return goal_definition


def BehaviorFromTrack(track, params, xy_offset, start, end):
    return BehaviorStaticTrajectory(params, TrajectoryFromTrack(track, xy_offset=xy_offset, start=start, end=end))


class InteractionDatasetReader:
    def __init__(self, **kwargs):
        self._track_dict_cache = {}

        self._use_shape_from_track = kwargs.pop("use_shape_from_track", True)
        self._use_rectangle_shape = kwargs.pop("use_rectangle_shape", True)
        self._wb = kwargs.pop("wb", 2.7)  # wheelbase
        self._crad = kwargs.pop("crad", 1.0)  # collision radius

    def TrackFromTrackfile(self, filename, track_id):
        if filename not in self._track_dict_cache:
            try:
                self._track_dict_cache[filename] = dataset_reader.read_tracks(
                    filename)
            except FileNotFoundError as e:
                logging.error("File {} not found!".format(
                    os.path.abspath(filename)))
                exit(1)
        track = self._track_dict_cache[filename][track_id]
        # TODO: Filter track
        return track

    def AgentFromTrackfile(self, track_params, param_server, scenario_track_info, agent_id, goal_def):

        if scenario_track_info.GetEgoTrackInfo().GetTrackId() == agent_id:
            agent_track_info = scenario_track_info.GetEgoTrackInfo()
        elif agent_id in scenario_track_info.GetOtherTrackInfos().keys():
            agent_track_info = scenario_track_info.GetOtherTrackInfos()[
                agent_id]
        else:
            raise ValueError("unknown agent id {}".format(agent_id))
        fname = agent_track_info.GetFileName()  # track_params["filename"]
        track_id = agent_track_info.GetTrackId()  # track_params["track_id"]
        agent_id = agent_track_info.GetTrackId()
        track = self.TrackFromTrackfile(fname, track_id)

        xy_offset = scenario_track_info.GetXYOffset()

        # create behavior model from track, we use start time of scenario here
        start_time = scenario_track_info.GetStartTimeMs()
        end_time = scenario_track_info.GetEndTimeMs()
        behavior_model = track_params["behavior_model"]
        model_converter = ModelJsonConversion()
        if behavior_model is None:
            # each agent need's its own param server
            behavior = BehaviorFromTrack(track, param_server.AddChild(
                "agent{}".format(agent_id)), xy_offset, start_time, end_time)
        else:
            behavior = behavior_model

        # retrieve initial state of valid agent
        if agent_id in scenario_track_info._other_agents_track_infos:
            start_time_init_state = max(
                scenario_track_info._other_agents_track_infos[agent_id].GetStartTimeMs(), start_time)
        else:
            start_time_init_state = start_time
        try:
            initial_state = InitStateFromTrack(
                track, xy_offset, start_time_init_state)
        except:
            raise ValueError("Could not retrieve initial state of agent {} at t={}.".format(
                agent_id, start_time_init_state))

        if self._use_shape_from_track:
            wb = WheelbaseFromTrack(track)
            crad = ColRadiusFromTrack(track)
        else:
            wb = self._wb
            crad = self._crad

        param_server["DynamicModel"]["wheel_base"] = wb
        try:
            dynamic_model = model_converter.convert_model(
                track_params["dynamic_model"], param_server)
        except:
            raise ValueError("Could not create dynamic_model")

        try:
            execution_model = model_converter.convert_model(
                track_params["execution_model"], param_server)
        except:
            raise ValueError("Could not retrieve execution_model")

        try:
            if self._use_rectangle_shape:
                vehicle_shape = GenerateCarRectangle(wb, crad)
            else:
                vehicle_shape = GenerateCarLimousine(wb, crad)
        except:
            raise ValueError("Could not create vehicle_shape")

        if goal_def is None:
            goal_def = GoalDefinitionFromTrack(
                track, end_time, xy_offset=xy_offset)

        bark_agent = Agent(
            initial_state,
            behavior,
            dynamic_model,
            execution_model,
            vehicle_shape,
            param_server.AddChild("agent{}".format(agent_id)),
            goal_def,
            track_params["map_interface"])
        # set agent id from track
        bark_agent.SetAgentId(track_id)
        return bark_agent



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import sys
import logging
import time
import tensorflow as tf
tf.compat.v1.enable_v2_behavior()

from tf_agents.drivers import dynamic_step_driver
from tf_agents.drivers import dynamic_episode_driver
from tf_agents.policies import actor_policy
from bark.runtime.commons.parameters import ParameterServer

from tf_agents.metrics import tf_metrics
from tf_agents.eval import metric_utils
from tf_agents.utils import common
from tf_agents.trajectories import time_step as ts

from bark_ml.library_wrappers.lib_tf_agents.runners.tfa_runner import TFARunner


class PPORunner(TFARunner):
  def __init__(self,
               environment=None,
               agent=None,
               params=None):
    TFARunner.__init__(self,
                       environment=environment,
                       agent=agent,
                       params=params)

  def _train(self):
    global_iteration = self._agent._agent._train_step_counter.numpy()
    for i in range(1, self._params["ML"]["PPORunner"]["NumberOfCollections", "", 10000]):
      tf.keras.backend.clear_session()
      print(f"Collection {i}")
      global_iteration = self._agent._agent._train_step_counter.numpy()
      tf.summary.experimental.set_step(global_iteration)
      self._collection_driver.run()
      trajectories = self._agent._replay_buffer.gather_all()
      self._agent._agent.train(experience=trajectories)
      self._agent._replay_buffer.clear()
      if i % self._params["ML"]["PPORunner"]["EvaluateEveryNSteps", "", 100] == 0:
        self._tracer.Reset()
        self.Run(
          num_episodes=self._params["ML"]["TFARunner"]["EvaluationSteps", "", 20],
          mode="training")
        self._agent.Save()


# coding=utf-8
# Copyright 2018 The TF-Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sample Keras actor network that generates distributions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gin
import numpy as np
import tensorflow as tf  # pylint: disable=g-explicit-tensorflow-version-import

from tf_agents.networks import categorical_projection_network
from tf_agents.networks import encoding_network
from tf_agents.networks import network
from tf_agents.networks import normal_projection_network
from tf_agents.specs import tensor_spec
from tf_agents.utils import nest_utils

from bark.runtime.commons.parameters import ParameterServer


def _categorical_projection_net(action_spec, logits_init_output_factor=0.1):
  return categorical_projection_network.CategoricalProjectionNetwork(
      action_spec, logits_init_output_factor=logits_init_output_factor)


def _normal_projection_net(action_spec,
                           init_action_stddev=0.35,
                           init_means_output_factor=0.1):
  std_bias_initializer_value = np.log(np.exp(init_action_stddev) - 1)

  return normal_projection_network.NormalProjectionNetwork(
      action_spec,
      init_means_output_factor=init_means_output_factor,
      std_bias_initializer_value=std_bias_initializer_value,
      scale_distribution=False)


@gin.configurable
class GNNActorDistributionNetwork(network.DistributionNetwork):
  """Creates an actor producing either Normal or Categorical distribution.

  Note: By default, this network uses `NormalProjectionNetwork` for continuous
  projection which by default uses `tanh_squash_to_spec` to normalize its
  output. Due to the nature of the `tanh` function, values near the spec bounds
  cannot be returned.
  """

  def __init__(self,
               input_tensor_spec,
               output_tensor_spec,
               gnn,
               preprocessing_layers=None,
               preprocessing_combiner=None,
               conv_layer_params=None,
               fc_layer_params=(200, 100),
               dropout_layer_params=None,
               activation_fn=tf.keras.activations.relu,
               kernel_initializer=None,
               batch_squash=False,
               dtype=tf.float32,
               discrete_projection_net=_categorical_projection_net,
               continuous_projection_net=_normal_projection_net,
               name='ActorDistributionNetwork',
               params=ParameterServer()):
    """Creates an instance of `ActorDistributionNetwork`.

    Args:
      input_tensor_spec: A nest of `tensor_spec.TensorSpec` representing the
        input.
      output_tensor_spec: A nest of `tensor_spec.BoundedTensorSpec` representing
        the output.
      preprocessing_layers: (Optional.) A nest of `tf.keras.layers.Layer`
        representing preprocessing for the different observations.
        All of these layers must not be already built. For more details see
        the documentation of `networks.EncodingNetwork`.
      preprocessing_combiner: (Optional.) A keras layer that takes a flat list
        of tensors and combines them. Good options include
        `tf.keras.layers.Add` and `tf.keras.layers.Concatenate(axis=-1)`.
        This layer must not be already built. For more details see
        the documentation of `networks.EncodingNetwork`.
      conv_layer_params: Optional list of convolution layers parameters, where
        each item is a length-three tuple indicating (filters, kernel_size,
        stride).
      fc_layer_params: Optional list of fully_connected parameters, where each
        item is the number of units in the layer.
      dropout_layer_params: Optional list of dropout layer parameters, each item
        is the fraction of input units to drop or a dictionary of parameters
        according to the keras.Dropout documentation. The additional parameter
        `permanent', if set to True, allows to apply dropout at inference for
        approximated Bayesian inference. The dropout layers are interleaved with
        the fully connected layers; there is a dropout layer after each fully
        connected layer, except if the entry in the list is None. This list must
        have the same length of fc_layer_params, or be None.
      activation_fn: Activation function, e.g. tf.nn.relu, slim.leaky_relu, ...
      kernel_initializer: Initializer to use for the kernels of the conv and
        dense layers. If none is provided a default glorot_uniform
      batch_squash: If True the outer_ranks of the observation are squashed into
        the batch dimension. This allow encoding networks to be used with
        observations with shape [BxTx...].
      dtype: The dtype to use by the convolution and fully connected layers.
      discrete_projection_net: Callable that generates a discrete projection
        network to be called with some hidden state and the outer_rank of the
        state.
      continuous_projection_net: Callable that generates a continuous projection
        network to be called with some hidden state and the outer_rank of the
        state.
      name: A string representing name of the network.

    Raises:
      ValueError: If `input_tensor_spec` contains more than one observation.
    """

    if not kernel_initializer:
      kernel_initializer = tf.compat.v1.keras.initializers.glorot_uniform()
    if gnn is None:
      raise ValueError('`gnn` must not be `None`.')

    self._gnn = gnn(name=name + "_GNN", params=params)
    
    encoder = encoding_network.EncodingNetwork(
        input_tensor_spec=tf.TensorSpec([None, self._gnn._embedding_size]),
        preprocessing_layers=preprocessing_layers,
        preprocessing_combiner=preprocessing_combiner,
        conv_layer_params=conv_layer_params,
        fc_layer_params=fc_layer_params,
        dropout_layer_params=dropout_layer_params,
        activation_fn=activation_fn,
        kernel_initializer=kernel_initializer,
        batch_squash=batch_squash,
        dtype=dtype)

    def map_proj(spec):
      if tensor_spec.is_discrete(spec):
        return discrete_projection_net(spec)
      else:
        return continuous_projection_net(spec)

    projection_networks = tf.nest.map_structure(map_proj, output_tensor_spec)
    output_spec = tf.nest.map_structure(lambda proj_net: proj_net.output_spec,
                                        projection_networks)

    super(GNNActorDistributionNetwork, self).__init__(
        input_tensor_spec=input_tensor_spec,
        state_spec=(),
        output_spec=output_spec,
        name=name)

    self._encoder = encoder
    self._projection_networks = projection_networks
    self._output_tensor_spec = output_tensor_spec

  @property
  def output_tensor_spec(self):
    return self._output_tensor_spec

  def call(self,
           observations,
           step_type,
           network_state,
           training=False,
           mask=None):
    
    if len(tf.shape(observations)) == 2 or len(tf.shape(observations)) == 1:
      observations = tf.reshape(observations, [1, -1])
    
    if len(tf.shape(observations)) == 3:
      observations = tf.squeeze(observations, axis=0)
    
    embeddings = self._gnn(observations, training=training)
    # extract ego state (node 0)
    # print(embeddings)
    
    if tf.shape(embeddings)[0] > 0:
      embeddings = embeddings[:, 0]
      
    with tf.name_scope("PPOActorNetwork"):
      tf.summary.histogram("embedding", embeddings)
    

    state, network_state = self._encoder(
      embeddings,
      step_type=step_type,
      network_state=network_state,
      training=training)
    
    outer_rank = nest_utils.get_outer_rank(observations, self.input_tensor_spec)

    def call_projection_net(proj_net):
      distribution, _ = proj_net(
          state, outer_rank, training=training, mask=mask)
      return distribution

    output_actions = tf.nest.map_structure(
        call_projection_net, self._projection_networks)
    # print(output_actions, "output_actions")
    return output_actions, network_state
  


# Copyright (c) 2021 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.
import math
import numpy as np
from bark.core.geometry import *
from bark.core.models.dynamic import *
from bark.core.world.opendrive import *
from bark.core.world.goal_definition import *
from bark.core.world.renderer import *
from bark.runtime.commons.parameters import ParameterServer


class BufferedViewer:
  def __init__(self, params=None, **kwargs):
    self._params = params or ParameterServer()
    self._vparams = self._params["Visualization"]
    # map color config
    self._dashed_sc = self._vparams[
      "DashedStrokeColor", "", [128, 128, 128, 64]] 
    self._dashed_sw = self._vparams["DashedStrokeWidth", "",  0.1]
    self._solid_sc = self._vparams[
      "SolidStrokeColor", "", [12, 44, 132, 128]]
    self._solid_sw = self._vparams["SolidStrokeWidth", "", 0.2]
    # agent color config
    self._ego_sc = self._vparams[
      "EgoStrokeColor", "", [34, 94, 168, 255]] 
    self._ego_fc = self._vparams[
      "EgoFaceColor", "", [34, 94, 168, 128]] 
    self._other_sc = self._vparams[
      "OtherStrokeColor", "", [128, 128, 128, 255]] 
    self._other_fc = self._vparams[
      "OtherFaceColor", "", [128, 128, 128, 128]] 
    self._agent_height = self._vparams[
      "AgentHeight", "", 2.] 
    
  def DrawAgents(self, world, eval_agent_id):
    ego_agent = world.agents[eval_agent_id]
    for agent in world.agents.values():
      agent_poly = agent.GetPolygonFromState(agent.state)
      agent_primitive = RenderPrimitive(agent_poly)
      agent_primitive.Add("agent_id", str(agent.id))
      agent_primitive.Add("height", self._agent_height)
      if agent == ego_agent:
        agent_primitive.Add("stroke_color", self._ego_sc)
        agent_primitive.Add("fill_color", self._ego_fc)
        world.renderer.Add("EGO_AGENT", agent_primitive)    
      else:
        agent_primitive.Add("stroke_color", self._other_sc)
        agent_primitive.Add("fill_color", self._other_fc)
        world.renderer.Add("OTHER_AGENT", agent_primitive)
      
  def drawWorld(self, world, eval_agent_ids=None, scenario_idx=None):
    self.DrawMap(world)
    self.DrawAgents(world, eval_agent_ids[0])
    # add ego state
    ego_agent = world.agents[eval_agent_ids[0]]
    ego_agent_state = ego_agent.state
    ego_vel_primitive = RenderPrimitive(ego_agent_state)
    world.renderer.Add("EGO_AGENT_STATE", ego_vel_primitive)
    # TODO: e.g., set map origin

  def DrawMap(self, world):
    map = world.map.GetOpenDriveMap()
    for road in map.GetRoads().values():
      for lane_section in road.lane_sections:
        for lane in lane_section.GetLanes().values():
          self.DrawLane(lane, world)
      
  def DrawLane(self, lane, world):
    lane_primitive = RenderPrimitive(lane.line)
    lane_primitive.Add("stroke_color", self._solid_sc)
    lane_primitive.Add("stroke_width", self._solid_sw)
    if lane.road_mark.type == XodrRoadMarkType.broken or \
      lane.road_mark.type == XodrRoadMarkType.none:
      lane_primitive.Add("stroke_color", self._dashed_sc)
      lane_primitive.Add("stroke_width", self._dashed_sw)
    world.renderer.Add("MAP_LINE", lane_primitive)

  # NOTE: legacy functions
  def clear(self):
    pass
  def Reset(self):
    pass



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from bark.core.world.agent import *
from bark.core.models.behavior import *
from bark.core.world import *
from bark.core.world.map import *
from bark.core.models.dynamic import *
from bark.core.models.execution import *
from bark.core.geometry import *
from bark.core.geometry.standard_shapes import *
from bark.runtime.commons import ParameterServer
import inspect
import re
import sys
import numpy as np


class ModelJsonConversion:

  def agent_from_json(self, agent_json, param_server):
    bark_agent = Agent(
      np.array(agent_json["state"]), 
      self.convert_model(agent_json["behavior_model"], param_server), 
      self.convert_model(agent_json["dynamic_model"], param_server),
      self.convert_model(agent_json["execution_model"], param_server), 
      Polygon2d(agent_json["shape"]["center_pose"],
      np.array(agent_json["shape"]["polygon_points"])),
      param_server.AddChild("agent"),
      agent_json["goal_definition"],
      agent_json["map_interface"])
    return bark_agent

  def agent_to_json(self, agent):
    agent_json = {}
    agent_json["behavior_model"] = self.convert_model(agent.behavior_model)
    agent_json["dynamic_model"] = self.convert_model(agent.dynamic_model)
    agent_json["execution_model"] = self.convert_model(agent.execution_model)
    agent_json["id"] = agent.id
    agent_json["state"] = agent.state.tolist()
    agent_json["shape"] = {}
    agent_json["shape"]["polygon_points"] = agent.shape.ToArray().tolist()
    agent_json["shape"]["center_pose"] = agent.shape.center.tolist()
    agent_json["followed_trajectory"] = agent.followed_trajectory.tolist()
    agent_json["planned_trajectory"] = agent.planned_trajectory.tolist()
    return agent_json

  def convert_model(self, model, params=None, dynamic_model=None):
    if isinstance(model,str):
      try:
        if params is None:
          return eval("{}()".format(model))
        else:
          try:
            return eval("{}(params)".format(model))    
          except:
            return eval("{}(dynamic_model, params)".format(model))                 
      except NameError:
        raise NameError("Unkown model type: {}".format(input))
    else:
        return str(model).split(".")[-1]

  def extract_models(self, module, filter_regexp):
    all_module_members = inspect.getmembers(sys.modules[module])
    model_list = []
    for module_member in all_module_members:
      member_name = module_member[0]
      m = re.match(filter_regexp,member_name)
      if m:
        model_list.append(m.group())
    return model_list



from bark_ml.library_wrappers.lib_tf_agents.networks.gnn_actor_network import GNNActorNetwork
from bark_ml.library_wrappers.lib_tf_agents.networks.gnn_actor_distribution_network import GNNActorDistributionNetwork
from bark_ml.library_wrappers.lib_tf_agents.networks.gnn_critic_network import GNNCriticNetwork
from bark_ml.library_wrappers.lib_tf_agents.networks.gnn_value_network import GNNValueNetwork


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import numpy as np

from bark.core.models.behavior import BehaviorModel, BehaviorDynamicModel
from bark_ml.commons.py_spaces import BoundedContinuous


class BehaviorContinuousML(BehaviorDynamicModel):
  def __init__(self,
               params=None):
    BehaviorDynamicModel.__init__(self, params)
    # BehaviorModel.__init__(self, params)
    self._lower_bounds = params["ML"]["BehaviorContinuousML"][
      "ActionsLowerBound",
      "Lower-bound for actions.",
      [-5.0, -0.2]]
    self._upper_bounds = params["ML"]["BehaviorContinuousML"][
      "ActionsUpperBound",
      "Upper-bound for actions.",
      [4.0, 0.2]]

  @property
  def action_space(self):
    return BoundedContinuous(
      2,  # acceleration and steering-rate
      low=np.array(self._lower_bounds, dtype=np.float32),
      high=np.array(self._upper_bounds, dtype=np.float32))
  
  def Clone(self):
    return self